from collections import Counter

import numpy as np

from multi_imbalance.resampling.static_smote import StaticSMOTE


def test_static_smote():
    X = np.vstack([np.random.normal(0, 1, (100, 2)),
                   np.random.normal(3, 5, (30, 2)),
                   np.random.normal(-2, 2, (20, 2))])

    y = np.array([1] * 100 + [2] * 30 + [3] * 20)
    ssm = StaticSMOTE()
    X_resampled, y_resampled = ssm.fit_resample(X, y)
    cnt = Counter(y_resampled)
    assert cnt[1] == 100
    assert cnt[2] == 60
    assert cnt[3] == 80



import unittest
from unittest.mock import MagicMock

from sklearn.tree import DecisionTreeClassifier

from multi_imbalance.ensemble.mrbbagging import MRBBagging
import numpy as np

X_train = np.array([
    [0.05837771, 0.57543339],
    [0.06153624, 0.99871925],
    [0.14308529, 0.00681144],
    [0.23401697, 0.21188708],
    [0.2418553, 0.02137086],
    [0.32480534, 0.81547632],
    [0.42478482, 0.31995162],
    [0.50726834, 0.72621157],
    [0.54580968, 0.58025914],
    [0.55748531, 0.71866238],
    [0.69208769, 0.63759459],
    [0.70797377, 0.16348051],
    [0.76410615, 0.70451542],
    [0.81680686, 0.50793884],
    [0.8490789, 0.53826627],
    [0.8847505, 0.96856011],
])
y_train = np.array([0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0])

X_test = np.array([[0.9287003, 0.97580299],
                   [0.9584236, 0.10536541],
                   [0.01, 0.87666093],
                   [0.97352367, 0.78807909], ])

y_test = np.array([0, 0, 0, 0])


class TestMRBBagging(unittest.TestCase):
    def test_api(self):
        mrbbagging = MRBBagging(1, DecisionTreeClassifier(random_state=0), random_state=0)
        mrbbagging.fit(X_train, y_train)
        y_pred = mrbbagging.predict(X_test)
        assert all(y_pred == y_test)

    def test_api_multiple_trees(self):
        mrbbagging = MRBBagging(5, DecisionTreeClassifier(random_state=0), random_state=0)
        mrbbagging.fit(X_train, y_train)
        y_pred = mrbbagging.predict(X_test)
        assert all(y_pred == y_test)

    def test_api_with_feature_selection(self):
        mrbbagging = MRBBagging(1, DecisionTreeClassifier(random_state=0), feature_selection=True, random_state=0)
        mrbbagging.fit(X_train, y_train)
        y_pred = mrbbagging.predict(X_test)
        assert all(y_pred == y_test)

    def test_api_with_random_feature_selection(self):
        mrbbagging = MRBBagging(1, DecisionTreeClassifier(random_state=0), feature_selection=True, random_fs=True,
                                random_state=0)
        mrbbagging.fit(X_train, y_train)
        y_pred = mrbbagging.predict(X_test)
        assert all(y_pred == y_test)

    def test_api_with_feature_selection_sqrt_features(self):
        mrbbagging = MRBBagging(1, DecisionTreeClassifier(random_state=0), feature_selection=True,
                                half_features=False, random_state=0)
        mrbbagging.fit(X_train, y_train)
        y_pred = mrbbagging.predict(X_test)
        assert all(y_pred == y_test)

    def test__group_data(self):
        mrbbagging = MRBBagging(1, DecisionTreeClassifier())
        x = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        y = ["A", "B", "C"]
        classes, grouped_data = mrbbagging._group_data(x, y)
        self.assertEqual(classes, {'A', 'B', 'C'})
        self.assertEqual(grouped_data, {'C': [[[3, 3, 3], 'C']], 'A': [[[1, 1, 1], 'A']], 'B': [[[2, 2, 2], 'B']]})

    def test__group_data_with_none(self):
        mrbbagging = MRBBagging(1, DecisionTreeClassifier())
        x = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        y = ["A", None, "C"]
        with self.assertRaises(AssertionError):
            mrbbagging._group_data(x, y)

    def test_fit_with_invalid_labels(self):
        mrbbagging = MRBBagging(1, MagicMock())
        x = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        y = ["A", "B"]
        with self.assertRaises(AssertionError):
            mrbbagging.fit(x, y)

    def test_fit_with_invalid_classifier(self):
        with self.assertRaises(AssertionError):
            MRBBagging(1, None)

    def test_with_invalid_k(self):
        with self.assertRaises(AssertionError):
            MRBBagging(0, DecisionTreeClassifier())


if __name__ == '__main__':
    unittest.main()



from collections import Counter

import numpy as np
import pytest

from multi_imbalance.resampling.global_cs import GlobalCS

X = np.array([
    [0.05837771, 0.57543339],
    [0.06153624, 0.99871925],
    [0.14308529, 0.00681144],
    [0.23401697, 0.21188708],
    [0.2418553, 0.02137086],
    [0.32480534, 0.81547632],
    [0.42478482, 0.31995162],
    [0.50726834, 0.72621157],
    [0.54580968, 0.58025914],
    [0.55748531, 0.71866238],
    [0.69208769, 0.63759459],
    [0.70797377, 0.16348051],
    [0.76410615, 0.70451542],
    [0.81680686, 0.50793884],
    [0.8490789, 0.53826627],
    [0.8847505, 0.96856011],
    [0.9287003, 0.97580299],
    [0.9584236, 0.10536541],
    [0.96983103, 0.87666093],
    [0.97352367, 0.78807909],
])

y_balanced = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
y_imb_easy = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])
y_imb_hard = np.array([0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0])
complete_test_data = [
    (X, y_balanced),
    (X, y_imb_easy),
    (X, y_imb_hard),
]


@pytest.fixture()
def global_cs_mock():
    def _get_parametrized_globalcs(X, y):
        clf = GlobalCS()
        clf.quantities = Counter(y)
        return clf

    return _get_parametrized_globalcs


def get_goal_quantity(y):
    quantities = Counter(y)
    return max(quantities.values()) * len(quantities.keys())


def calc_duplicates_quantities(X, y, X_oversampled):
    quantities = dict()
    for label in Counter(y).keys():
        quantities[label] = list()

    for i, row in enumerate(X):
        equal_row_indices = np.where((X_oversampled == row).all(axis=1))[0]
        label = y[i]
        quantities[label].append(len(equal_row_indices))

    return quantities


@pytest.mark.parametrize("X, y", complete_test_data)
def test_output_length_validate(X, y, global_cs_mock):
    clf = global_cs_mock(X, y)
    oversampled_X, oversampled_y = clf.fit_resample(X, y)
    assert len(oversampled_X) == get_goal_quantity(y)
    assert len(oversampled_y) == get_goal_quantity(y)


@pytest.mark.parametrize("X, y", complete_test_data)
def test_output_equal_replication(X, y, global_cs_mock):
    clf = global_cs_mock(X, y)
    oversampled_X, oversampled_y = clf.fit_resample(X, y)
    oversampled_class_quantities = calc_duplicates_quantities(X, y, oversampled_X)

    for label, duplicates_quantities in oversampled_class_quantities.items():
        min_quantity = min(duplicates_quantities)
        max_quantity = max(duplicates_quantities)

        assert max_quantity - min_quantity <= 1



from collections import Counter

import numpy as np

from multi_imbalance.resampling.spider import SPIDER3
from multi_imbalance.utils.array_util import (union, intersect, setdiff)

cost = np.ones((3, 3))
np.fill_diagonal(cost, 0)
spider = SPIDER3(1, maj_int_min={'maj': ["MAJ"], 'int': ["INT"], 'min': ["MIN"]}, cost=cost)


def test_union():
    arr1 = np.array([[1, 2, 3]])
    arr2 = np.array([[4, 5, 6]])
    actual = union(arr1, arr2)
    expected = np.array([[1, 2, 3], [4, 5, 6]])
    assert (actual == expected).all()

    arr1 = np.array([[1, 2, 3], [4, 5, 6]])
    arr2 = np.array([[1, 2, 3]])

    actual = union(arr1, arr2)
    expected = np.array([[1, 2, 3], [4, 5, 6], [1, 2, 3]])

    assert (actual == expected).all()


def test_intersect():
    arr1 = np.array([[1, 2, 3], [4, 5, 6]])
    arr2 = np.array([[1, 2, 3]])

    actual = intersect(arr1, arr2)
    expected = np.array([[1, 2, 3]])

    assert (actual == expected).all()


def test_setdiff():
    arr1 = np.array([[1, 2, 3], [4, 5, 6]])
    arr2 = np.array([[1, 2, 3]])

    actual = setdiff(arr1, arr2)
    expected = np.array([[4, 5, 6]])

    assert (actual == expected).all()


def test_knn():
    X = np.array([
        [1, 1],
        [1, -1],
        [-1, 1],
        [-1, -1],
        [0, 0]
    ]).astype(object)

    y = np.array(["MIN", "MIN", "MAJ", "MAJ", "MAJ"])

    DS = np.append(X, y.reshape(y.shape[0], 1), axis=1)

    assert (DS[4] == spider._knn(DS[0], DS)).all()


def test_min_cost_classes():
    X = np.array([
        [1, 1],
        [1, -1],
        [-1, 1],
        [-1, -1],
        [0, 0]
    ]).astype(object)

    y = np.array(["MIN", "MIN", "MAJ", "MAJ", "MAJ"])

    DS = np.append(X, y.reshape(y.shape[0], 1), axis=1)
    spider._fit_resample(X, y)
    spider._min_cost_classes(DS[0], DS)
    assert (spider._min_cost_classes(DS[0], DS) == ["MAJ"]).all()
    assert (spider._min_cost_classes(DS[4], DS) == ["MIN", "MAJ"]).all()

    
def test_estimate_cost_matrix():
    y = [0, 1, 1, 2, 2, 2, 2, 2, 2]
    cost = SPIDER3._estimate_cost_matrix(y).ravel().tolist()
    assert cost == [0, 1, 1, 2, 0, 1, 6, 3, 0]


def test_fit_resample():
    np.random.seed(7)
    X = np.vstack([np.random.normal(0, 1, (100, 2)),
                   np.random.normal(3, 5, (30, 2)),
                   np.random.normal(-2, 2, (20, 2))])

    y = np.array([1] * 100 + [2] * 30 + [3] * 20)
    sp = SPIDER3(5)
    X_resampled, y_resampled = sp.fit_resample(X, y)
    cnt = Counter(y_resampled)
    assert cnt[1] == 72
    assert cnt[2] == 57
    assert cnt[3] == 30


import os
from collections import Counter
from collections import defaultdict
from copy import deepcopy

import numpy as np
from imblearn.over_sampling import SMOTE
from sklearn.ensemble import BaggingClassifier
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import check_random_state

from multi_imbalance.resampling.global_cs import GlobalCS
from multi_imbalance.resampling.soup import SOUP


class ECOC(BaggingClassifier):
    """
    ECOC (Error Correcting Output Codes) is ensemble method for multi-class classification problems.
    Each class is encoded with unique binary or ternary code (where 0 means that class is excluded from training set
    of binary classifier). Then in the learning phase each binary classifier is learned. In the decoding phase the class
    which is closest to test instance in the sense of Hamming distance is chosen.
    """

    _allowed_encodings = ['dense', 'sparse', 'complete', 'OVA', 'OVO']
    _allowed_oversampling = [None, 'globalCS', 'SMOTE', 'SOUP']
    _allowed_classifiers = ['tree', 'NB', 'KNN']
    _allowed_weights = [None, 'acc', 'avg_tpr_min']

    def __init__(self, binary_classifier='KNN', preprocessing='SOUP', encoding='OVO', n_neighbors=3,
                 weights=None):
        """
        :param binary_classifier:
            binary classifier used by the algorithm. Possible classifiers:

            * 'tree':
                Decision Tree Classifier,
            * 'NB':
                Naive Bayes Classifier,
            * 'KNN' :
                K-Nearest Neighbors
            * 'ClassifierMixin' :
                An instance of a class that implements ClassifierMixin
        :param preprocessing:
            method for oversampling between aggregated classes in each dichotomy. Possible methods:

            * None :
                no oversampling applied,
            * 'globalCS' :
                random oversampling - randomly chosen instances of minority classes are duplicated
            * 'SMOTE' :
                Synthetic Minority Oversampling Technique
            * 'SOUP' :
                Similarity Oversampling Undersampling Preprocessing
            * 'TransformerMixin' :
                An instance of a class that implements TransformerMixin
        :param encoding:
            algorithm for encoding classes. Possible encodings:

            * 'dense':
                ceil(10log2(num_of_classes)) dichotomies, -1 and 1 with probability 0.5 each
            * 'sparse' :
                ceil(10log2(num_of_classes)) dichotomies, 0 with probability 0.5, -1 and 1 with probability 0.25 each
            * 'OVO' :
                'one vs one' - n(n-1)/2 dichotomies, where n is number of classes, one for each pair of classes.
                Each column has one 1 and one -1 for classes included in particular pair, 0s for remaining classes.
            * 'OVA' :
                'one vs all' - number of dichotomies is equal to number of classes. Each column has one 1 and
                -1 for all remaining rows
            * 'complete' : 2^(n-1)-1 dichotomies, reference
                T. G. Dietterich and G. Bakiri.
                Solving multiclass learning problems via error-correcting output codes.
                Journal of Artificial Intelligence Research, 2:263–286, 1995.
        :param n_neighbors:
        :param weights:
            strategy for dichotomies weighting. Possible values:

            * None :
                no weighting applied
            * 'acc' :
                accuracy-based weights
            * 'avg_tpr_min' :
                weights based on average true positive rates of dichotomies
        """
        super().__init__()
        self.binary_classifier = binary_classifier
        self.encoding = encoding
        self.preprocessing = preprocessing
        self.n_neighbors = n_neighbors
        self.weights = weights

        self.minority_classes = list()

        self._code_matrix = None
        self._binary_classifiers = []
        self._labels = None
        self._dich_weights = None

    def fit(self, X, y, minority_classes=None):
        """

        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :param y:
            one dimensional numpy array with labels for rows in X
        :param minority_classes:
            list of classes considered to be minority classes
        :return:
            self: object
        """
        if minority_classes is not None:
            self.minority_classes = minority_classes

        if self.weights is not None:
            X_train, X_for_weights, y_train, y_for_weights = train_test_split(X, y, test_size=0.2, stratify=y,
                                                                              random_state=0)
        else:
            X_train, y_train = X, y

        self._labels = np.unique(y)
        self._gen_code_matrix()
        self._binary_classifiers = [self._get_classifier() for _ in range(self._code_matrix.shape[1])]
        self._learn_binary_classifiers(X_train, y_train)
        if self.weights is not None:
            self._calc_weights(X_for_weights, y_for_weights)
        return self

    def predict(self, X):
        """
        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :return:
            numpy array, shape = [number of samples]. Predicted target values for X.
        """
        output_codes = np.zeros((X.shape[0], self._code_matrix.shape[1]))
        for classifier_idx, classifier in enumerate(self._binary_classifiers):
            output_codes[:, classifier_idx] = classifier.predict(X)

        predicted = np.zeros(X.shape[0])
        for row_idx, encoded_row in enumerate(output_codes):
            predicted[row_idx] = self._get_closest_class(encoded_row)

        return predicted

    def _learn_binary_classifiers(self, X, y):
        for classifier_idx, classifier in enumerate(self._binary_classifiers):
            excluded_classes_indices = [idx for idx in range(len(y)) if
                                        self._code_matrix[self._labels.tolist().index(y[idx])][classifier_idx] == 0]
            X_filtered = np.delete(X, excluded_classes_indices, 0)
            y_filtered = np.delete(y, excluded_classes_indices)
            binary_labels = np.array([self._code_matrix[self._labels.tolist().index(clazz)][classifier_idx] for clazz in
                                      y_filtered])
            X_filtered, binary_labels = self._oversample(X_filtered, binary_labels)
            classifier.fit(X_filtered, binary_labels)

    def _gen_code_matrix(self):
        if self.encoding == 'dense':
            self._code_matrix = self._encode_dense(self._labels.shape[0])
        elif self.encoding == 'sparse':
            self._code_matrix = self._encode_sparse(self._labels.shape[0])
        elif self.encoding == 'complete':
            self._code_matrix = self._encode_complete(self._labels.shape[0])
        elif self.encoding == 'OVO':
            self._code_matrix = self._encode_ovo(self._labels.shape[0])
        elif self.encoding == 'OVA':
            self._code_matrix = self._encode_ova(self._labels.shape[0])
        else:
            raise ValueError("Unknown matrix generation encoding: %s, expected to be one of %s."
                             % (self.encoding, ECOC._allowed_encodings))

    def _encode_dense(self, number_of_classes, random_state=0, number_of_code_generations=10000):
        try:
            dirname = os.path.dirname(__file__)
            matrix = np.load(dirname + f'/cached_matrices/dense_{number_of_classes}.npy')
            return matrix
        except IOError:
            print(f'Could not find cached matrix for dense code for {number_of_classes} classes, generating matrix...')

        number_of_columns = int(np.ceil(10 * np.log2(number_of_classes)))
        code_matrix = np.ones((number_of_classes, number_of_columns))
        random_state = check_random_state(random_state)

        max_min_dist = 0
        for i in range(number_of_code_generations):
            tmp_code_matrix = np.ones((number_of_classes, number_of_columns))
            min_dist = float('inf')

            for row in range(0, number_of_classes):
                for col in range(0, number_of_columns):
                    if random_state.randint(0, 2) == 1:
                        tmp_code_matrix[row, col] = -1

                for compared_row in range(0, row):
                    dist = self._hamming_distance(tmp_code_matrix[compared_row], tmp_code_matrix[row])
                    if dist < min_dist:
                        min_dist = dist

            if min_dist > max_min_dist:
                max_min_dist = min_dist
                code_matrix = tmp_code_matrix
        return code_matrix

    def _encode_sparse(self, number_of_classes, random_state=0, number_of_code_generations=10000):
        try:
            dirname = os.path.dirname(__file__)
            matrix = np.load(dirname + f'/cached_matrices/sparse_{number_of_classes}.npy')
            return matrix
        except IOError:
            print(f'Could not find cached matrix for sparse code for {number_of_classes} classes, generating matrix...')

        number_of_columns = int(np.ceil(15 * np.log2(number_of_classes)))
        code_matrix = np.ones((number_of_classes, number_of_columns))
        random_state = check_random_state(random_state)

        max_min_dist = 0
        for i in range(number_of_code_generations):
            tmp_code_matrix = np.ones((number_of_classes, number_of_columns))
            min_dist = float('inf')

            for row in range(0, number_of_classes):
                for col in range(0, number_of_columns):
                    rand = random_state.randint(0, 4)
                    if rand < 2:
                        tmp_code_matrix[row, col] = 0
                    elif rand == 3:
                        tmp_code_matrix[row, col] = -1

                if np.count_nonzero(tmp_code_matrix[row]) == 0:
                    break

                for compared_row in range(0, row):
                    dist = self._hamming_distance(tmp_code_matrix[compared_row], tmp_code_matrix[row])
                    if dist < min_dist:
                        min_dist = dist

            if self._has_matrix_all_zeros_column(tmp_code_matrix):
                continue

            if min_dist > max_min_dist:
                max_min_dist = min_dist
                code_matrix = tmp_code_matrix

        return code_matrix

    def _encode_ova(self, number_of_classes):
        matrix = np.identity(number_of_classes)
        matrix[matrix == 0] = -1
        return matrix

    def _encode_ovo(self, number_of_classes):
        number_of_columns = int(number_of_classes * (number_of_classes - 1) / 2)
        matrix = np.zeros((number_of_classes, number_of_columns), dtype=int)
        indices_map = self._map_indices_to_class_pairs(number_of_classes)
        for row in range(number_of_classes):
            for col in range(number_of_columns):
                if indices_map[col][0] == row:
                    matrix[row, col] = 1
                elif indices_map[col][1] == row:
                    matrix[row, col] = -1
        return matrix

    def _map_indices_to_class_pairs(self, number_of_classes):
        indices_map = dict()
        idx = 0
        for i in range(number_of_classes):
            for j in range(i + 1, number_of_classes):
                indices_map[idx] = (i, j)
                idx += 1
        return indices_map

    def _encode_complete(self, number_of_classes):
        code_length = 2 ** (number_of_classes - 1) - 1
        matrix = np.ones((number_of_classes, code_length))
        for row_idx in range(1, number_of_classes):
            digit = -1
            partial_code_len = 2 ** (number_of_classes - row_idx - 1)
            for idx in range(0, code_length, partial_code_len):
                matrix[row_idx][idx:idx + partial_code_len] = digit
                digit *= -1
        return matrix

    def _hamming_distance(self, v1, v2):
        return np.count_nonzero(v1 != v2)

    def _has_matrix_all_zeros_column(self, matrix):
        return (~matrix.any(axis=0)).any()

    def _get_closest_class(self, row):
        if self.weights is not None:
            return self._labels[
                np.argmin(
                    [sum(np.multiply(self.dich_weights, (encoded_class - row) ** 2)) for encoded_class in
                     self._code_matrix])]
        else:
            return self._labels[
                np.argmin([self._hamming_distance(row, encoded_class) for encoded_class in self._code_matrix])]

    def _oversample(self, X, y):
        if self.preprocessing is None:
            return X, y

        if isinstance(self.preprocessing, str):
            if self.preprocessing not in ECOC._allowed_oversampling:
                raise ValueError("Unknown preprocessing method: %s, expected to be one of %s."
                                 % (self.preprocessing, ECOC._allowed_oversampling))
            elif np.unique(y).size == 1:
                return X, y
            elif self.preprocessing == 'globalCS':
                gcs = GlobalCS()
                return gcs.fit_resample(X, y)
            elif self.preprocessing == 'SMOTE':
                return self._smote_oversample(X, y)
            elif self.preprocessing == 'SOUP':
                soup = SOUP()
                return soup.fit_resample(X, y)
        else:
            if not hasattr(self.preprocessing, 'fit_transform'):
                raise ValueError("Your resampler must implement fit_transform method")
            return self.preprocessing.fit_transform(X, y)

    def _get_classifier(self):
        if isinstance(self.binary_classifier, str):
            if self.binary_classifier not in ECOC._allowed_classifiers:
                raise ValueError(
                    "Unknown binary classifier: %s, expected to be one of %s."
                    % (self.binary_classifier, ECOC._allowed_classifiers))
            elif self.binary_classifier == 'tree':
                decision_tree_classifier = DecisionTreeClassifier(random_state=42)
                return decision_tree_classifier
            elif self.binary_classifier == 'NB':
                gnb = GaussianNB()
                return gnb
            elif self.binary_classifier == 'KNN':
                knn = KNeighborsClassifier(n_neighbors=self.n_neighbors)
                return knn
        else:
            if not hasattr(self.binary_classifier, 'fit') or not hasattr(self.binary_classifier, 'predict'):
                raise ValueError("Your classifier must implement fit and predict methods")
            return deepcopy(self.binary_classifier)

    def _smote_oversample(self, X, y):
        n_neighbors = min(3, min(np.unique(y, return_counts=True)[1]) - 1)
        if n_neighbors == 0:
            raise ValueError(
                'In order to use SMOTE preprocessing, the training set should contain at least 2 examples from each class')
        smote = SMOTE(k_neighbors=n_neighbors, random_state=42)
        return smote.fit_resample(X, y)

    def _calc_weights(self, X_for_weights, y_for_weights):
        if self.weights not in ECOC._allowed_weights:
            raise ValueError("Unknown weighting strategy: %s, expected to be one of %s."
                             % (self.weights, ECOC._allowed_weights))

        dich_weights = np.ones(self._code_matrix.shape[1])
        if self.weights == 'acc':
            for clf_idx, clf in enumerate(self._binary_classifiers):
                samples_no = 0
                correct_no = 0
                for sample, sample_label in zip(X_for_weights, y_for_weights):
                    if self._code_matrix[np.where(self._labels == sample_label)[0][0]][clf_idx] != 0:
                        samples_no += 1
                        if clf.predict([sample])[0] == \
                                self._code_matrix[np.where(self._labels == sample_label)[0][0]][clf_idx]:
                            correct_no += 1
                if samples_no != 0:
                    acc = correct_no / samples_no
                    dich_weights[clf_idx] = -1 + 2 * acc
        elif self.weights == 'avg_tpr_min':
            min_counter = Counter([y for y in y_for_weights if y in self.minority_classes])

            for clf_idx, clf in enumerate(self._binary_classifiers):
                min_correct_pred = defaultdict(lambda: 0)
                for sample, sample_label in zip(X_for_weights, y_for_weights):
                    if clf.predict([sample])[0] == \
                            self._code_matrix[np.where(self._labels == sample_label)[0][0]][clf_idx]:
                        min_correct_pred[sample_label] += 1
                avg_tpr_min = np.mean([min_correct_pred[clazz] / min_counter[clazz] for clazz in min_counter.keys()])
                dich_weights[clf_idx] = avg_tpr_min

        self.dich_weights = dich_weights



import pytest
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

import multi_imbalance.ensemble.ecoc as ecoc

X = np.array([
    [1.8938566839198983, 0.7347724642028586, 1.5817290619305417],
    [1.6893330472771877, 1.3729481360429043, 0.1779576959347715],
    [1.1103882804642866, 0.2684931500114267, 0.24565717871603532],
    [0.9635120154904986, 0.44338438370111577, 1.6559238383999697],
    [0.6525827502237067, 0.8978087724631425, 1.5056794207545134],
    [0.8232009732859464, 0.5270243940630088, 1.434695372722657],
    [0.519304726338536, 0.4635228434262648, 0.014170648565480004],
    [1.3938520002157688, 1.524670776643407, 0.9011189423913637],
    [0.09993454781831534, 0.5991188594563008, 0.6462181194010983],
    [1.5300019511124079, 0.08177359763506553, 1.7642527715349894],
    [1.1770688242955876, 0.9604049547799067, 0.6989025594835503],
    [1.5143651712498534, 1.4914673103908214, 1.3377704178955587],
    [1.1299009013495136, 0.700540900007983, 1.071829181951729],
    [1.530652133805449, 0.2992536048983532, 1.957731948975865],
    [1.6236761570974148, 0.5919033806975751, 1.6334065904199757],
    [0.9365056250644108, 1.526475631725099, 1.420298571686271],
    [0.9063995770780813, 1.0248369545634513, 1.36911505163145],
    [0.3861789635773656, 0.5758917834278445, 0.910187724154228],
    [0.7165380621896438, 1.494299618627891, 0.521854931610239],
    [1.3621939213912113, 0.387219837127391, 1.321376123618781],
    [1.6764775993219296, 0.15364096535456317, 1.3219739817389],
    [3.6764775993219296, 0.15364096535456317, 1.3789731273891],
    [4.6764775993219296, 2.15364096535456317, 1.9830281123211],
    [5.6764775993219296, 3.15364096535456317, 1.3213121322321],
    [6.321312, 11.15364096535456317, 1.0998908320132],
    [8.414132131, 2.15364096535456317, 5.0998908320132],
    [10.6764775993219296, -2.15364096535456317, 3.0998908320132],
    [4.6764775993219296, 0.15364096535456317, 2.0998908320132],
    [-4.6764775993219296, -1.15364096535456317, 9.0998908320132],
    [-6.6764775993219296, 11.15364096535456317, 5.0998908320132],
])

y = np.array([2, 0, 2, 3, 0, 3, 1, 0, 2, 0, 2, 3, 1, 2, 1, 3, 0, 3, 2, 0, 0, 1, 2, 3, 0, 1, 2, 3, 1, 2])


def test_random_oversampling():
    ecoc_clf = ecoc.ECOC(preprocessing='globalCS')
    X_oversampled, y_oversampled = ecoc_clf._oversample(X, y)

    assert len(X_oversampled) == len(y_oversampled)
    assert len(set(np.unique(y_oversampled, return_counts=True)[1])) == 1
    assert set(y_oversampled).issubset(set(y))


def test_no_oversampling():
    ecoc_clf = ecoc.ECOC(preprocessing=None)
    X_oversampled, y_oversampled = ecoc_clf._oversample(X, y)

    assert X.shape == X_oversampled.shape
    assert y.shape == y_oversampled.shape


@pytest.mark.parametrize("encoding_strategy", ['dense', 'sparse', 'OVO', 'OVA', 'complete'])
@pytest.mark.parametrize("oversampling", [None, 'globalCS', 'SMOTE', 'SOUP'])
def test_encoding(encoding_strategy, oversampling):
    ecoc_clf = ecoc.ECOC(encoding=encoding_strategy, preprocessing=oversampling)
    ecoc_clf.fit(X, y)
    matrix = ecoc_clf._code_matrix

    number_of_classes = len(np.unique(y))

    assert matrix.shape[0] == number_of_classes
    assert len(np.unique(matrix, axis=0)) == number_of_classes
    assert bool((~matrix.any(axis=0)).any()) is False


@pytest.mark.parametrize("encoding_strategy", ['dense', 'sparse'])
def test_dense_and_sparse_with_not_cached_matrices(encoding_strategy):
    X1 = np.concatenate((X, 2 * X, 3 * X, 4 * X, 5 * X), axis=0)
    y1 = np.concatenate((y + 4, y + 8, y + 12, y + 16, y + 20))

    ecoc_clf = ecoc.ECOC(encoding=encoding_strategy)
    ecoc_clf.fit(X1, y1)
    matrix = ecoc_clf._code_matrix

    number_of_classes = len(np.unique(y1))

    assert matrix.shape[0] == number_of_classes
    assert len(np.unique(matrix, axis=0)) == number_of_classes
    assert bool((~matrix.any(axis=0)).any()) is False


def test_hamming_distance():
    v1 = np.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1])
    v2 = np.array([-1, 1, -1, 1, -1, 0, 1, 0, 1, 0, -1, -1, -1, -1])
    distance = ecoc.ECOC()._hamming_distance(v1, v2)

    assert distance == 5


def test_with_own_classifier():
    class DummyClassifier:
        def fit(self, X, y):
            pass

        def predict(self, X):
            return np.zeros(len(X))

    dummy_clf = DummyClassifier()
    ecoc_clf = ecoc.ECOC(binary_classifier=dummy_clf, preprocessing=None)
    ecoc_clf.fit(X, y)
    predicted = ecoc_clf.predict(np.array([[1.0, 2.0], [4.0, 5.5], [6.7, 8.8]]))
    assert np.all(predicted == 0)


def test_with_own_preprocessing():
    class DummyResampler:
        def fit_transform(self, X, y):
            return np.concatenate((X, X), axis=0), np.concatenate((y, y), axis=None)

    dummy_resampler = DummyResampler()
    ecoc_clf = ecoc.ECOC(preprocessing=dummy_resampler)
    X_oversampled, y_oversampled = ecoc_clf._oversample(X, y)
    assert len(X_oversampled) == 2 * len(X)
    assert len(y_oversampled) == 2 * len(y)


def test_unknown_classifier():
    ecoc_clf = ecoc.ECOC(binary_classifier='DUMMY_CLASSIFIER', preprocessing=None)
    with pytest.raises(ValueError) as e:
        ecoc_clf.fit(X, y)
    assert 'DUMMY_CLASSIFIER' in str(e.value)


def test_own_classifier_without_predict_and_fit():
    class DummyClassifier:
        def foo(self, X, y):
            pass

        def bar(self, X):
            return np.zeros(len(X))

    dummy_clf = DummyClassifier()
    ecoc_clf = ecoc.ECOC(binary_classifier=dummy_clf, preprocessing=None)
    with pytest.raises(ValueError) as e:
        ecoc_clf.fit(X, y)
    assert 'predict' in str(e.value)
    assert 'fit' in str(e.value)


@pytest.mark.parametrize("classifier", ['tree', 'NB', 'KNN'])
@pytest.mark.parametrize("weights", [None, 'acc', 'avg_tpr_min'])
def test_predefined_classifiers_and_weighting_without_exceptions(classifier, weights):
    ecoc_clf = ecoc.ECOC(binary_classifier=classifier, weights=weights)
    ecoc_clf.fit(X, y)
    predicted = ecoc_clf.predict(np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]))
    assert len(predicted) == 3


def test_unknown_preprocessing():
    ecoc_clf = ecoc.ECOC(preprocessing='DUMMY_OVERSAMPLING')
    with pytest.raises(ValueError) as e:
        ecoc_clf.fit(X, y)
    assert 'DUMMY_OVERSAMPLING' in str(e.value)


def test_own_preprocessing_without_fit_transform():
    class DummyOversampler:
        def foo(self, X, y):
            pass

        def bar(self, X):
            return np.zeros(len(X))

    dummy_oversampler = DummyOversampler()
    ecoc_clf = ecoc.ECOC(preprocessing=dummy_oversampler)
    with pytest.raises(ValueError) as e:
        ecoc_clf.fit(X, y)
    assert 'fit_transform' in str(e.value)


@pytest.mark.parametrize("encoding_strategy", ['dense', 'sparse', 'OVO', 'OVA', 'complete'])
@pytest.mark.parametrize("oversampling", [None, 'globalCS', 'SMOTE', 'SOUP'])
def test_ecoc_with_sklearn_pipeline(encoding_strategy, oversampling):
    pipeline = Pipeline([
        ('scaler', StandardScaler()),
        ('ecoc', ecoc.ECOC(encoding=encoding_strategy, preprocessing=oversampling))
    ])
    pipeline.fit(X, y)
    y_hat = pipeline.predict(np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]))
    assert len(y_hat) == 3



from imblearn.metrics import geometric_mean_score


def gmean_score(y_test, y_pred, correction: float = 0.001) -> float:  # pragma no cover
    """
    Calculate geometric mean score

    :param y_test:
        numpy array with labels
    :param y_pred:
        numpy array with predicted labels
    :param correction:
        value that replaces 0 during multiplication to avoid zeroing the result
    :return:
        geometric_mean_score: float
    """
    return geometric_mean_score(y_test, y_pred, correction=correction)



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils import resample
from sklearn.utils.validation import check_is_fitted

from multi_imbalance.ensemble.soup_bagging import SOUPBagging

X_train = np.array([
    [0.05837771, 0.57543339],
    [0.06153624, 0.99871925],
    [0.14308529, 0.00681144],
    [0.23401697, 0.21188708],
    [0.2418553, 0.02137086],
    [0.32480534, 0.81547632],
    [0.42478482, 0.31995162],
    [0.50726834, 0.72621157],
    [0.54580968, 0.58025914],
    [0.55748531, 0.71866238],
    [0.69208769, 0.63759459],
    [0.70797377, 0.16348051],
    [0.76410615, 0.70451542],
    [0.81680686, 0.50793884],
    [0.8490789, 0.53826627],
    [0.8847505, 0.96856011],
])
y_train = np.array([0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0])

X_test = np.array([[0.9287003, 0.97580299],
                   [0.9584236, 0.10536541],
                   [0.01, 0.87666093],
                   [0.97352367, 0.78807909], ])

y_test = np.array([0, 1, 0, 0])


def test_soubagging():
    clf = KNeighborsClassifier()
    maj_int_min = {'maj': [0], 'int': [], 'min': [1]}
    clf = SOUPBagging(clf, n_classifiers=2, maj_int_min=maj_int_min)
    clf.fit(X_train, y_train)
    y_pred = clf.predict(X_test)
    assert all(y_pred == y_test)
    y_pred = clf.predict(X_test, strategy='mixed')
    assert all(y_pred == y_test)
    y_pred = clf.predict(X_test, strategy='optimistic')
    assert all(y_pred == y_test)
    y_pred = clf.predict(X_test, strategy='pessimistic')
    assert all(y_pred == y_test)
    y_pred = clf.predict(X_test, strategy='global')
    assert all(y_pred == y_test)


def test_exception():
    clf = KNeighborsClassifier()
    maj_int_min = {'maj': [0], 'int': [], 'min': [1]}
    clf = SOUPBagging(clf, n_classifiers=2, maj_int_min=maj_int_min)
    clf.fit(X_train, y_train)
    with pytest.raises(KeyError):
        clf.predict(X_test, 'incorrect')


def test_default_classifier():
    maj_int_min = {'maj': [0], 'int': [], 'min': [1]}
    clf = SOUPBagging(n_classifiers=2, maj_int_min=maj_int_min)
    clf.fit(X_train, y_train)
    y_pred = clf.predict(X_test)
    assert all(y_pred == y_test)


def test_fit_classifier_classifier():
    maj_int_min = {'maj': [0], 'int': [], 'min': [1]}
    clf = KNeighborsClassifier()
    clf, weights = SOUPBagging.fit_classifier([clf, X_train, y_train,
                                               resample(X_train, y_train, stratify=y_train, random_state=0),
                                               maj_int_min])
    y_pred = clf.predict(X_test)

    check_is_fitted(clf)
    assert all(y_pred == y_test)
    assert_array_almost_equal(weights, np.array([1.33288904, 0.66644452]))




import numpy as np


def setdiff(arr1, arr2):
    """
    Performs the difference over two numpy arrays.

    :param arr1:
        Numpy array number 1.
    :param arr2:
        Numpy array number 2.
    :return:
        Result of the difference of arr1 and arr2.
    """

    for element in arr2:
        if contains(arr1, element):
            arr1 = np.delete(arr1, index_of(arr1, element), 0)
    return arr1


def union(arr1, arr2):
    """
    Performs the union over two numpy arrays
    (not removing duplicates, as it's how the algorithm SPIDER3 actually works).

    :param arr1:
        Numpy array number 1.
    :param arr2:
        Numpy array number 2.
    :return:
        The union of arr1 and arr2.
    """

    if arr1.size == 0:
        return arr2
    elif arr2.size == 0:
        return arr1
    else:
        return np.append(arr1, arr2, axis=0)


def contains(dataset, example):
    """
    Returns if dataset contains the example.
    :param dataset:
    :param example:
    :return: True or False depending on whether dataset contains the example.
    """
    for x in dataset:
        if all(x == example):
            return True
    return False


def index_of(arr, example):
    """
    :return: Index of learning exmaple in arr.
    """
    for i, x in enumerate(arr):
        if all(x == example):
            return i
    return -1


def intersect(arr1, arr2):
    """
    Performs the intersection operation over two numpy arrays (not removing duplicates).

    :param arr1:
        Numpy array number 1.
    :param arr2:
        Numpy array number 2.
    :return:
        The intersection of arr1 and arr2.
    """

    if arr1.size == 0 or arr2.size == 0:
        return np.array([])

    result = np.array([])
    for x in arr1:
        if contains(arr2, x):
            result = union(result, np.array([x]))
    return result



import pytest
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

import multi_imbalance.ensemble.ovo as ovo
import numpy as np

X = np.array([
    [-0.5813674466943386, -0.37091887120486655, -0.4465813355321204],
    [-0.630844420005455, 0.2871060228285258, 0.25613448374582437],
    [-0.6714353752038125, -0.3537255703996809, -0.9687281557330454],
    [0.48996953214789785, -0.09697447345439447, 0.5495667841083927],
    [-0.8821146485100975, -0.7739441502933209, 0.34906417794620515],
    [-0.6652132165510964, -0.8488383805882527, -0.030511639438375093],
    [0.7846621478367604, -0.9231479370667406, 0.7262231362586529],
    [0.7860907554630845, -0.33615224298146584, 0.6928271619140047],
    [0.7630774674537872, -0.7753044382704197, -0.7570971821030896],
    [-0.5843764899573332, -0.524996569658353, 0.9675951634125524],
    [-33.5843764899573332, -0.303030303030303, 5.9292929292929290],
    [22.5843764899573332, -2.020202022020202, 6.3213211113213211],
    [11.5843764899573332, -22.110101010011010, 7.3213123131232111],
    [2.5843764899573332, -1.010123211231232, 1.9675951634125524],
    [-0.5843764899573332, 1.321321312112321, 2.3213123123123222],
    [-0.2321312313211321, -412.321321312112321, 6.1010101010100101],
    [-0.3921809321038213, -2.321321312112321, 3.1010101010100101],
    [1.5843764899573332, -0.4324234243242342, 32.1010101010100101],
    [2.5843764899573332, -0.321321312112321, 53.1010101010100101],
    [42.5843764899573332, 66.321321312112321, 3242.1010101010100101],
    [1.5843764899573332, 44.321321312112321, 2.999909909090909090],
    [-2.5843764899573332, 12.321321312112321, 2112.342423],
    [-5.5843764899573332, 3.321321312112321, 2212.1010101010100101],
    [-8.5843764899573332, 1.321321312112321, 1222.1010101010100101],
    [-2.3213123213211232, -2.321321312112321, 992.1010101010100101],
])

y = np.array([1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 1, 1, 2, 3, 2, 1, 2, 3, 1, 2, 3, 1, 2, 3, 1])


def test_fit_predict():
    clf = ovo.OVO(preprocessing=None)
    clf.fit(X[:-1], y[:-1])
    predicted = clf.predict(X[-1].reshape(1, -1))
    assert predicted in (1, 2, 3)


@pytest.mark.parametrize("classifier", ['tree', 'NB', 'KNN'])
def test_binary_classifiers(classifier):
    clf = ovo.OVO(binary_classifier=classifier, preprocessing=None)
    clf.fit(X[:-1], y[:-1])
    classifiers = clf._binary_classifiers

    assert len(classifiers) == 3
    assert classifiers[1][0].predict(X[-1].reshape(1, -1)) in (1, 2)
    assert classifiers[2][0].predict(X[-1].reshape(1, -1)) in (1, 3)
    assert classifiers[2][1].predict(X[-1].reshape(1, -1)) in (2, 3)


def test_max_voting():
    labels = np.array([4, 3, 6, 5, 7])
    binary_outputs = np.array([[0, 0, 0, 0, 0],
                               [3, 0, 0, 0, 0],
                               [4, 3, 0, 0, 0],
                               [4, 5, 6, 0, 0],
                               [7, 7, 7, 5, 0]])

    clf = ovo.OVO()
    clf._labels = labels
    voting_winner = clf._perform_max_voting(binary_outputs)
    assert voting_winner == 7


def test_with_own_classifier():
    class DummyClassifier:
        def fit(self, X, y):
            pass

        def predict(self, X):
            return np.ones(len(X))

    dummy_clf = DummyClassifier()
    ovo_clf = ovo.OVO(binary_classifier=dummy_clf, preprocessing=None)
    ovo_clf.fit(X, y)
    predicted = ovo_clf.predict(np.array([[1.0, 2.0], [4.0, 5.5], [6.7, 8.8]]))
    assert np.all(predicted == 1)


def test_with_own_preprocessing():
    class DummyResampler:
        def fit_resample(self, X, y):
            return np.concatenate((X, X), axis=0), np.concatenate((y, y), axis=None)

    dummy_resampler = DummyResampler()
    ovo_clf = ovo.OVO(preprocessing=dummy_resampler)
    X_oversampled, y_oversampled = ovo_clf._oversample(X, y)
    assert len(X_oversampled) == 2 * len(X)
    assert len(y_oversampled) == 2 * len(y)


@pytest.mark.parametrize("preprocessing_btwn", ['all', 'maj-min'])
@pytest.mark.parametrize("classifier", ['tree', 'NB', 'KNN'])
@pytest.mark.parametrize("preprocessing", [None, 'globalCS', 'SMOTE', 'SOUP'])
def test_predefined_classifiers_and_preprocessings_without_errors(classifier, preprocessing, preprocessing_btwn):
    ovo_clf = ovo.OVO(binary_classifier=classifier, preprocessing=preprocessing,
                      preprocessing_between=preprocessing_btwn)
    ovo_clf.fit(X, y)
    predicted = ovo_clf.predict(np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]))
    assert len(predicted) == 3


def test_unknown_preprocessing():
    ovo_clf = ovo.OVO(preprocessing='DUMMY_OVERSAMPLING')
    with pytest.raises(ValueError) as e:
        ovo_clf.fit(X, y)
    assert 'DUMMY_OVERSAMPLING' in str(e.value)


def test_own_preprocessing_without_fit_resample():
    class DummyOversampler:
        def foo(self, X, y):
            pass

        def bar(self, X):
            return np.zeros(len(X))

    dummy_oversampler = DummyOversampler()
    ovo_clf = ovo.OVO(preprocessing=dummy_oversampler)
    with pytest.raises(ValueError) as e:
        ovo_clf.fit(X, y)
    assert 'fit_resample' in str(e.value)


def test_unknown_preprocessing_between_strategy_raises_exception():
    ovo_clf = ovo.OVO(preprocessing_between='min-intermediate')
    with pytest.raises(ValueError) as e:
        ovo_clf.fit(X, y)
    assert 'min-intermediate' in str(e.value)


@pytest.mark.parametrize("preprocessing_btwn", ['all', 'maj-min'])
@pytest.mark.parametrize("classifier", ['tree', 'NB', 'KNN'])
@pytest.mark.parametrize("preprocessing", [None, 'globalCS', 'SMOTE', 'SOUP'])
def test_ecoc_with_sklearn_pipeline(preprocessing_btwn, classifier, preprocessing):
    pipeline = Pipeline([
        ('scaler', StandardScaler()),
        ('ecoc', ovo.OVO(binary_classifier=classifier, preprocessing=preprocessing,
                         preprocessing_between=preprocessing_btwn))
    ])
    pipeline.fit(X, y)
    y_hat = pipeline.predict(np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], [7.7, 8.8, 9.99]]))
    assert len(y_hat) == 3



"""Collection of imbalanced datasets.

The characteristics of the available datasets are presented in the table below.

 ID    Name             Repository     Class distribution                                            #S       #F
 1     balance_scale    UCI            46.1 : 46.1 : 7.8                                             625      4
 2     cmc              UCI            42.7 : 34.7 : 22.6                                            1,473    9
 3     cleveland        UCI            53.9 : 18.2 : 11.8 : 11.8 : 4.4                               297      13
 4     dermatology      UCI            31   : 19.8 : 16.8 : 13.4 : 13.4 : 5.6                        358      34
 5     ecoli            UCI            42.6 : 22.9 : 15.5 : 10.4 : 6    : 1.5  : 0.6 : 0.6           336      7
 6     glass            UCI            35.5 : 32.7 : 13.6 : 8    : 6.1  : 4.2                        214      9
 7     hayes_roth       UCI            38.6 : 38.6 : 22.7                                            132      5
 8     new_thyroid      UCI            69.8 : 16.3 : 14                                              215      5
 9     winequailty_red  UCI            42.6 : 39.9 : 12.4 : 3.3  : 1.1 : 0.6                         1599     11
 10    yeast            UCI            31.2 : 28.9 : 16.4 : 11   : 3.4 : 3 : 2.5 : 2 : 1.4 : 0.3    1,484    9
"""
# TODO add missing characteristics

from collections import OrderedDict
import tarfile
from io import BytesIO
from os import makedirs
from os.path import join, isfile

import numpy as np

from sklearn.datasets._base import Bunch

PRE_FILENAME = 'x'
POST_FILENAME = 'data.npz'
DATA_HOME_BASIC = "./../../data/"

MAP_NAME_ID_KEYS = ['1czysty-cut', '2delikatne-cut', '3mocniej-cut', '4delikatne-bezover-cut',
                    'balance-scale', 'cleveland', 'cleveland_v2', 'cmc', 'dermatology',
                    'glass', 'hayes-roth', 'new_ecoli', 'new_led7digit', 'new_vehicle',
                    'new_winequality-red', 'new_yeast', 'thyroid-newthyroid']

MAP_NAME_ID = OrderedDict()
MAP_ID_NAME = OrderedDict()
for v, k in enumerate(MAP_NAME_ID_KEYS):
    MAP_NAME_ID[k] = v + 1
    MAP_ID_NAME[v + 1] = k


def load_datasets(data_home=DATA_HOME_BASIC):
    """
    Load the benchmark datasets.

    :param data_home: Default catalogue in which the data is stored in .tar.gz format.
    :returns:
        OrderedDict of Bunch object. Each Bunch object refered as dataset have the following attributes:

            * dataset.data :
                ndarray, shape (n_samples, n_features)
            * dataset.target :
                ndarray, shape (n_samples, )
            * dataset.DESCR :
                string Description of the each dataset.
    """
    extracted_dir = join(data_home, "extracted")
    datasets = OrderedDict()

    filter_data_ = MAP_NAME_ID.keys()

    for it in filter_data_:
        filename = PRE_FILENAME + str(MAP_NAME_ID[it]) + POST_FILENAME
        filename = join(extracted_dir, filename)
        available = isfile(filename)

        if not available:
            makedirs(extracted_dir, exist_ok=True)
            with open(f'{data_home}data.tar.gz', 'rb') as fin:
                f = BytesIO(fin.read())
            tar = tarfile.open(fileobj=f)
            tar.extractall(path=extracted_dir)

        data = np.load(filename)
        X, y = data['data'], data['label']

        datasets[it] = Bunch(data=X, target=y, DESCR=it)

    return datasets



from collections import Counter

import numpy as np
import sklearn
from imblearn.base import BaseSampler


class GlobalCS(BaseSampler):
    """
    Global CS is an algorithm that equalizes number of samples in each class. It duplicates all samples equally
    for each class to achieve majority class size
    """

    def __init__(self, shuffle: bool = True):
        super().__init__()
        self._sampling_type = 'over-sampling'
        self.shuffle = shuffle
        self.quantities, self.max_quantity, self.X, self.y = [None] * 4

    def _fit_resample(self, X, y):
        """
        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :param y:
            one dimensional numpy array with labels for rows in X
        :return:
            Resampled X (max class quantity * number of unique classes), y (number of rows in X) as numpy array
        """
        assert len(X.shape) == 2, 'X should have 2 dimension'
        assert X.shape[0] == y.shape[0], 'Number of labels must be equal to number of samples'

        self.quantities = Counter(y)
        self.max_quantity = int(np.max(list(self.quantities.values())))
        self.X = X
        self.y = y

        result_X, result_y = list(), list()

        for class_name, class_quantity in self.quantities.items():
            temp_X, temp_y = self._equal_oversample(self.X, self.y, class_name)

            result_X.extend(temp_X)
            result_y.extend(temp_y)

        if self.shuffle:
            result_X, result_y = sklearn.utils.shuffle(result_X, result_y)

        return np.array(result_X), np.array(result_y)

    def _equal_oversample(self, X, y, class_name):
        indices_in_class = [i for i, class_label in enumerate(y) if class_label == class_name]
        desired_quantity = self.max_quantity - len(indices_in_class)

        oversampled_X, oversampled_y = list(X[indices_in_class]), list(y[indices_in_class])

        for i in range(desired_quantity):
            sample_index_to_duplicate: int = i % self.quantities[class_name]
            sample_id: int = indices_in_class[sample_index_to_duplicate]
            oversampled_X.append(X[sample_id])
            oversampled_y.append(y[sample_id])

        return oversampled_X, oversampled_y



from collections import Counter, defaultdict
from copy import deepcopy
from operator import itemgetter

import numpy as np
import sklearn
from imblearn.base import BaseSampler
from sklearn.neighbors import NearestNeighbors

from multi_imbalance.utils.data import construct_maj_int_min


class SOUP(BaseSampler):
    """
    Similarity Oversampling and Undersampling Preprocessing (SOUP) is an algorithm that equalizes number of samples
    in each class. It also takes care of the similarity between classes, which means that it removes samples from
    majority class, that are close to samples from the other class and duplicate samples from the minority classes,
    which are in the safest area in space
    """

    def __init__(self, k: int = 7, shuffle=False, maj_int_min=None) -> None:
        """
        :param k:
            number of neighbors
        :param shuffle:
            bool - output will be shuffled
        :param maj_int_min:
            dict {'maj': majority class labels, 'min': minority class labels}
        """
        super().__init__()
        self._sampling_type = 'clean-sampling'
        self.k = k
        self.shuffle = shuffle
        self.maj_int_min = maj_int_min
        self.quantities, self.goal_quantity = None, None
        self.dsc_maj_cls, self.asc_min_cls = None, None
        self._X, self._y = None, None

    def _fit_resample(self, X, y):
        """
        The method computes the metrics required for resampling based on the given set

        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :param y:
            one dimensional numpy array with labels for rows in X
        :return:
            Resampled X (median class quantity * number of unique classes), y (number of rows in X) as numpy array
        """

        if self.maj_int_min is None:
            self.maj_int_min = construct_maj_int_min(y)

        self._X = deepcopy(X)
        self._y = deepcopy(y)

        assert len(self._X.shape) == 2, 'X should have 2 dimension'
        assert self._X.shape[0] == self._y.shape[0], 'Number of labels must be equal to number of samples'

        self.quantities = Counter(self._y)
        self.goal_quantity = self._calculate_goal_quantity(self.maj_int_min)
        self.dsc_maj_cls = sorted(((v, i) for v, i in self.quantities.items() if i >= self.goal_quantity),
                                  key=itemgetter(1), reverse=True)
        self.asc_min_cls = sorted(((v, i) for v, i in self.quantities.items() if i < self.goal_quantity),
                                  key=itemgetter(1), reverse=False)

        for class_name, class_quantity in self.dsc_maj_cls:
            self._X, self._y = self._undersample(self._X, self._y, class_name)

        for class_name, class_quantity in self.asc_min_cls:
            self._X, self._y = self._oversample(self._X, self._y, class_name)

        if self.shuffle:
            self._X, self._y = sklearn.utils.shuffle(self._X, self._y)

        return np.array(self._X), np.array(self._y)

    def _construct_class_safe_levels(self, X, y, class_name) -> defaultdict:
        self.quantities = Counter(y)
        indices_in_class = [i for i, value in enumerate(y) if value == class_name]

        neigh_clf = NearestNeighbors(n_neighbors=self.k + 1).fit(X)
        neighbour_indices = neigh_clf.kneighbors(X[indices_in_class], return_distance=False)[:, 1:]
        neighbour_classes = y[neighbour_indices]

        class_safe_levels = defaultdict(float)
        for i, sample_id in enumerate(indices_in_class):
            neighbours_quantities = Counter(neighbour_classes[i])
            class_safe_levels[sample_id] = self._calculate_sample_safe_level(class_name, neighbours_quantities)

        return class_safe_levels

    def _calculate_sample_safe_level(self, class_name, neighbours_quantities: Counter):
        safe_level = 0
        q: Counter = self.quantities

        for neigh_label, neigh_q in neighbours_quantities.items():
            similarity_between_classes = min(q[class_name], q[neigh_label]) / max(q[class_name], q[neigh_label])
            safe_level += neigh_q * similarity_between_classes

        safe_level /= self.k

        if safe_level > 1:
            raise ValueError(f'Safe level is bigger than 1: {safe_level}')

        return safe_level

    def _undersample(self, X, y, class_name):
        safe_levels_of_samples_in_class = self._construct_class_safe_levels(X, y, class_name)

        class_quantity = self.quantities[class_name]
        safe_levels_list = sorted(safe_levels_of_samples_in_class.items(), key=itemgetter(1))
        samples_to_remove_quantity = max(0, int(class_quantity - self.goal_quantity))
        if samples_to_remove_quantity > 0:
            remove_indices = list(map(itemgetter(0), safe_levels_list[:samples_to_remove_quantity]))
            X = np.delete(X, remove_indices, axis=0)
            y = np.delete(y, remove_indices, axis=0)

        return X, y

    def _oversample(self, X, y, class_name):
        safe_levels_of_samples_in_class = self._construct_class_safe_levels(X, y, class_name)
        class_quantity = self.quantities[class_name]
        safe_levels_list = list(sorted(safe_levels_of_samples_in_class.items(), key=itemgetter(1), reverse=True))

        difference = self.goal_quantity - class_quantity
        while difference > 0:
            quantity_items_to_copy = min(difference, class_quantity)
            indices_to_copy = list(map(itemgetter(0), safe_levels_list[:quantity_items_to_copy]))
            X = np.vstack((X, X[indices_to_copy]))
            y = np.hstack((y, y[indices_to_copy]))
            difference -= quantity_items_to_copy

        return X, y

    def _calculate_goal_quantity(self, maj_int_min=None):
        if maj_int_min is None:
            maj_q = max(list(self.quantities.values()))
            min_q = min(list(self.quantities.values()))
            return np.mean((min_q, maj_q), dtype=int)
        else:
            maj_classes = {k: v for k, v in self.quantities.items() if k in maj_int_min['maj']}
            maj_q = list(maj_classes.values())
            min_classes = {k: v for k, v in self.quantities.items() if k in maj_int_min['min']}
            min_q = list(min_classes.values())

            if len(maj_q) == 0:
                return np.mean(min_q, dtype=int)
            if len(min_q) == 0:
                return np.mean(maj_q, dtype=int)

            return np.mean((max(min_q), min(maj_q)), dtype=int)



from collections import Counter, defaultdict

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from multi_imbalance.resampling.soup import SOUP

X = np.array([
    [0.05837771, 0.57543339],
    [0.06153624, 0.99871925],
    [0.14308529, 0.00681144],
    [0.23401697, 0.21188708],
    [0.2418553, 0.02137086],
    [0.32480534, 0.81547632],
    [0.42478482, 0.31995162],
    [0.50726834, 0.72621157],
    [0.54580968, 0.58025914],
    [0.55748531, 0.71866238],
    [0.69208769, 0.63759459],
    [0.70797377, 0.16348051],
    [0.76410615, 0.70451542],
    [0.81680686, 0.50793884],
    [0.8490789, 0.53826627],
    [0.8847505, 0.96856011],
    [0.9287003, 0.97580299],
    [0.9584236, 0.10536541],
    [0.96983103, 0.87666093],
    [0.97352367, 0.78807909],
])

y_balanced = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
y_balanced_first_sample_safe_level = 0.8
y_balanced_0_class_safe_levels = defaultdict(float,
                                             {0: 1.0, 1: 1.0, 2: 1.0, 3: 1.0, 4: 1.0, 5: 1.0, 6: 1.0, 7: 1.0, 8: 1.0,
                                              9: 1.0})
y_balanced_1_class_safe_levels = defaultdict(float,
                                             {10: 1.0, 11: 1.0, 12: 1.0, 13: 1.0, 14: 1.0, 15: 1.0, 16: 1.0, 17: 1.0,
                                              18: 1.0, 19: 1.0})

y_imb_easy = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])
y_imb_easy_first_sample_safe_level = 0.685714
y_imb_easy_0_class_safe_levels = defaultdict(float, {0: 1.0, 1: 1.0, 2: 1.0, 3: 1.0, 4: 1.0, 5: 1.0, 6: 1.0, 7: 1.0,
                                                     8: 0.8857142857142858, 9: 1.0, 10: 0.7714285714285714,
                                                     11: 0.7714285714285714, 12: 0.6571428571428571,
                                                     17: 0.7714285714285714})
y_imb_easy_1_class_safe_levels = defaultdict(float,
                                             {13: 0.6571428571428571, 14: 0.6571428571428571, 15: 0.7714285714285714,
                                              16: 0.7714285714285714, 18: 0.8857142857142858, 19: 0.8857142857142858})

y_imb_hard = np.array([0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0])
y_imb_hard_first_sample_safe_level = 0.685714
y_imb_hard_quantities_0_class_safe_levels = defaultdict(float, {0: 0.8857142857142858, 1: 0.7714285714285714,
                                                                2: 0.8857142857142858, 3: 0.8857142857142858,
                                                                4: 0.8857142857142858, 5: 0.7714285714285714,
                                                                7: 0.7714285714285714, 10: 0.6571428571428571,
                                                                11: 0.6571428571428571, 12: 0.7714285714285714,
                                                                13: 0.7714285714285714, 15: 0.7714285714285714,
                                                                17: 0.7714285714285714, 19: 0.6571428571428571})
y_imb_hard_quantities_1_class_safe_levels = defaultdict(float, {6: 0.5428571428571429, 8: 0.5428571428571429,
                                                                9: 0.5428571428571429, 14: 0.5428571428571429,
                                                                16: 0.5428571428571429, 18: 0.6571428571428571})

complete_test_data = [
    (X, y_balanced, y_balanced_0_class_safe_levels, y_balanced_1_class_safe_levels, y_balanced_first_sample_safe_level),
    (X, y_imb_easy, y_imb_easy_0_class_safe_levels, y_imb_easy_1_class_safe_levels, y_imb_easy_first_sample_safe_level),
    (X, y_imb_hard, y_imb_hard_quantities_0_class_safe_levels, y_imb_hard_quantities_1_class_safe_levels,
     y_imb_hard_first_sample_safe_level),
]

safe_levels_test_data = [
    # x,y,class_name,expected undersampling,oversampling quantity
    (X, y_balanced, 0, 20, 20),
    (X, y_balanced, 1, 20, 20),
    (X, y_imb_easy, 0, 16, 20),
    (X, y_imb_easy, 1, 20, 24),
    (X, y_imb_hard, 0, 16, 20),
    (X, y_imb_hard, 1, 20, 24),
]


@pytest.fixture()
def soup_mock():
    def _get_parametrized_soup(X, y):
        clf = SOUP(k=5)
        clf.quantities = Counter(y)
        clf.goal_quantity = clf._calculate_goal_quantity()
        return clf

    return _get_parametrized_soup


@pytest.mark.parametrize("X, y, zero_safe_levels, one_safe_levels, first_sample_safe", complete_test_data)
def test_calculating_safe_levels_for_sample(X, y, zero_safe_levels, one_safe_levels, first_sample_safe, soup_mock):
    clf = soup_mock(X, y)
    neighbour_quantities = Counter({0: 3, 1: 1})

    safe_level = clf._calculate_sample_safe_level(0, neighbour_quantities)
    assert_array_almost_equal(safe_level, first_sample_safe)


@pytest.mark.parametrize("X, y, zero_safe_levels, one_safe_levels, first_sample_safe", complete_test_data)
def test_calculating_safe_levels_for_class(X, y, zero_safe_levels, one_safe_levels, first_sample_safe,
                                           soup_mock):
    clf = soup_mock(X, y)

    zero_levels = clf._construct_class_safe_levels(X, y, 0)
    one_levels = clf._construct_class_safe_levels(X, y, 1)

    zero_levels == zero_safe_levels
    one_levels == one_safe_levels


@pytest.mark.parametrize("X, y, class_name, expected_undersampling, expected_oversampling", safe_levels_test_data)
def test_oversample(X, y, class_name, expected_undersampling, expected_oversampling, soup_mock):
    clf = soup_mock(X, y)
    oversampled_X, oversampled_y = clf._oversample(X, y, class_name)
    assert len(oversampled_X) == expected_oversampling
    assert len(oversampled_y) == expected_oversampling


@pytest.mark.parametrize("X, y, class_name, expected_undersampling, expected_oversampling", safe_levels_test_data)
def test_undersample(X, y, class_name, expected_undersampling, expected_oversampling, soup_mock):
    clf = soup_mock(X, y)
    undersampled_X, undersampled_y = clf._undersample(X, y, class_name)
    assert len(undersampled_X) == expected_undersampling
    assert len(undersampled_y) == expected_undersampling



from collections import Counter

import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.decomposition import PCA

from multi_imbalance.utils.data import construct_flat_2pc_df

sns.set_style('darkgrid')


def plot_cardinality_and_2d_data(X, y, dataset_name='') -> None:  # pragma no cover
    """
    Plots cardinality of classes from y as well as scatter plot of X transformed to two dimensions using PCA

    :param ndarray X:
        two dimensional numpy array
    :param ndarray y:
        one dimensional numpy array
    :param str dataset_name:
        title of chart
    """
    n = len(Counter(y).keys())
    p = sns.color_palette("husl", n)

    pca = PCA(n_components=2)
    pca.fit(X)

    fig, axs = plt.subplots(ncols=2)
    fig.set_size_inches(12, 6)
    axs = axs.flatten()

    fig.suptitle(dataset_name, fontsize=16)

    sns.countplot(y, ax=axs[0], palette=p)
    X = pca.transform(X)
    df = construct_flat_2pc_df(X, y)
    sns.scatterplot(x='x1', y='x2', hue='y', style='y', data=df, alpha=0.7, ax=axs[1], legend='full', palette=p)

    axs[0].set_xlabel("class")
    axs[0].set_ylabel("cardinality")


def plot_visual_comparision_datasets(X1, y1, X2, y2, dataset_name1='', dataset_name2='') -> None:  # pragma no cover
    """
    Plots comparision of X1 y1 and X2 y2 using plot_cardinality_and_2d_data, which plots cardinality of classes from
    y as well as scatter plot of X transformed to two dimensions using PCA

    :param ndarray X1:
        two dimensional numpy array with data from dataset1
    :param ndarray y1:
        one dimensional numpy array with target classes from dataset1
    :param ndarray X2:
        two dimensional numpy array with data from dataset2
    :param ndarray y2:
        one dimensional numpy array with target classes from dataset1
    :param str dataset_name1:
        first dataset chart title
    :param str dataset_name2:
        second dataset chart title
    """
    n = len(Counter(y1).keys())
    p = sns.color_palette("husl", n)

    pca = PCA(n_components=2)
    pca.fit(X1)

    fig, axs = plt.subplots(ncols=2, nrows=2)
    fig.set_size_inches(16, 10)
    axs = axs.flatten()

    sns.countplot(y1, ax=axs[0], palette=p)
    transformed_X = pca.transform(X1)
    df = construct_flat_2pc_df(transformed_X, y1)
    sns.scatterplot(x='x1', y='x2', hue='y', style='y', data=df, alpha=0.7, ax=axs[1], legend='full', palette=p)

    sns.countplot(y2, ax=axs[2], palette=p)
    transformed_X2 = pca.transform(X2)
    df = construct_flat_2pc_df(transformed_X2, y2)
    sns.scatterplot(x='x1', y='x2', hue='y', style='y', data=df, alpha=0.7, ax=axs[3], legend='full', palette=p)

    axs[0].set_xlabel("class")
    axs[2].set_xlabel("class")
    axs[0].set_ylabel("cardinality")
    axs[2].set_ylabel("cardinality")
    axs[1].set_title(dataset_name1)
    axs[3].set_title(dataset_name2)

    y_lim = axs[1].get_ylim()
    x_lim = axs[1].get_xlim()
    axs[3].set_ylim(y_lim)
    axs[3].set_xlim(x_lim)



"""Test the datasets loader.
"""

from multi_imbalance.datasets import load_datasets

DATASET_SHAPE = {
    '1czysty-cut': (1200, 2),
    '2delikatne-cut': (1200, 2),
    '3mocniej-cut': (1200, 2),
    '4delikatne-bezover-cut': (1200, 2),
    'balance-scale': (625, 4),
    'cleveland': (303, 13),
    'cleveland_v2': (303, 13),
    'cmc': (1473, 9),
    'dermatology': (366, 34),
    'glass': (214, 9),
    'hayes-roth': (160, 4),
    'new_ecoli': (336, 7),
    'new_led7digit': (500, 7),
    'new_vehicle': (846, 18),
    'new_winequality-red': (1599, 11),
    'new_yeast': (1484, 8),
    'thyroid-newthyroid': (215, 5)
}


def test_load_datasets():
    print("Testing loading datasets")
    datasets = load_datasets(data_home="./data/")
    for k in DATASET_SHAPE.keys():
        X = datasets[k].data
        assert DATASET_SHAPE[k] == X.shape



from copy import deepcopy

import numpy as np
from imblearn.over_sampling import SMOTE
from sklearn.ensemble import BaggingClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier

from multi_imbalance.resampling.global_cs import GlobalCS
from multi_imbalance.resampling.soup import SOUP


class OVO(BaggingClassifier):
    """

    OVO (One vs One) is an ensemble method that makes predictions for multi-class problems. OVO decomposes problem
    into m(m-1)/2 binary problems, where m is number of classes. Each of binary classifiers distinguishes between two
    classes. In the learning phase each classifier is learned only with instances from particular two classes. In
    prediction phase each classifier decides between these two classes. Results are aggregated and final output is
    derived depending on chosen aggregation model.

    """

    _allowed_classifiers = ['tree', 'NB', 'KNN']
    _allowed_preprocessing = [None, 'globalCS', 'SMOTE', 'SOUP']
    _allowed_preprocessing_between = ['all', 'maj-min']

    def __init__(self, binary_classifier='tree', n_neighbors=3, preprocessing='SOUP', preprocessing_between='all'):
        """
        :param binary_classifier:
            binary classifier. Possible classifiers:

            * 'tree':
                Decision Tree Classifier,
            * 'KNN':
                K-Nearest Neighbors
            * 'NB' :
                Naive Bayes
            * 'ClassifierMixin' :
                An instance of a class that implements ClassifierMixin
        :param n_neighbors:
            number of nearest neighbors in KNN, works only if binary_classifier=='KNN'
        :param preprocessing:
            method for preprocessing of pairs of classes in the learning phase of ensemble. Possible values:

            * None:
                no preprocessing applied
            * 'globalCS':
                oversampling with globalCS algorithm
            * 'SMOTE':
                oversampling with SMOTE algorithm
            * 'SOUP':
                oversampling and undersampling with SOUP algorithm
            * 'TransformerMixin' :
                An instance of a class that implements TransformerMixin
        :param preprocessing_between:
            types of classes between which resampling should be applied. Possible values:

            * 'all' :
                oversampling between each pair of classes
            * 'maj-min' :
                oversampling only between majority ad minority classes
        """
        super().__init__()
        self.binary_classifier = binary_classifier
        self.n_neighbors = n_neighbors
        self.preprocessing = preprocessing
        self.oversample_between = preprocessing_between
        self._binary_classifiers = []
        self._labels = np.array([])
        self._minority_classes = list()

    def fit(self, X, y, minority_classes=None):
        """
        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :param y:
            one dimensional numpy array with labels for rows in X
        :param minority_classes:
            list of classes considered to be minority
        :return:
            self: object
        """
        if minority_classes is None:
            minority_classes = list()

        self._labels = np.unique(y)
        self._minority_classes = minority_classes
        num_of_classes = len(self._labels)
        self._binary_classifiers = [[self._get_classifier() for _ in range(n)] for n in
                                    range(0, num_of_classes)]
        self._learn_binary_classifiers(X, y)
        return self

    def predict(self, X):
        """
        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :return:
            numpy array, shape = [number of samples]. Predicted target values for X.
        """
        num_of_classes = len(self._labels)
        predicted = list()
        for instance in X:
            binary_outputs_matrix = self._construct_binary_outputs_matrix(instance, num_of_classes)
            predicted.append(self._perform_max_voting(binary_outputs_matrix))

        return np.array(predicted)

    def _construct_binary_outputs_matrix(self, instance, num_of_classes):
        binary_outputs_matrix = np.zeros((num_of_classes, num_of_classes))
        for class_idx1 in range(len(self._labels)):
            for class_idx2 in range(class_idx1):
                binary_outputs_matrix[class_idx1][class_idx2] = self._binary_classifiers[class_idx1][class_idx2] \
                    .predict([instance])
        return binary_outputs_matrix

    def _learn_binary_classifiers(self, X, y):
        for row in range(len(self._labels)):
            for col in range(row):
                first_class, second_class = self._labels[row], self._labels[col]
                filtered_indices = [idx for idx in range(len(y)) if y[idx] in (first_class, second_class)]
                X_filtered, y_filtered = X[filtered_indices], y[filtered_indices]
                if self.should_perform_oversampling(first_class, second_class):
                    X_filtered, y_filtered = self._oversample(X_filtered, y_filtered)
                self._binary_classifiers[row][col].fit(X_filtered, y_filtered)

    def _get_classifier(self):
        if isinstance(self.binary_classifier, str):
            if self.binary_classifier not in OVO._allowed_classifiers:
                raise ValueError(
                    "Unknown binary classifier: %s, expected to be one of %s."
                    % (self.binary_classifier, OVO._allowed_classifiers))
            elif self.binary_classifier == 'tree':
                decision_tree_classifier = DecisionTreeClassifier(random_state=42)
                return decision_tree_classifier
            elif self.binary_classifier == 'NB':
                gnb = GaussianNB()
                return gnb
            elif self.binary_classifier == 'KNN':
                knn = KNeighborsClassifier(n_neighbors=self.n_neighbors)
                return knn
        else:
            if not hasattr(self.binary_classifier, 'fit') or not hasattr(self.binary_classifier, 'predict'):
                raise ValueError("Your classifier must implement fit and predict methods")
            return deepcopy(self.binary_classifier)

    def _perform_max_voting(self, binary_outputs_matrix):
        scores = np.zeros(len(self._labels))
        for clf_1 in range(len(binary_outputs_matrix)):
            for clf_2 in range(clf_1):
                scores[self._labels.tolist().index(binary_outputs_matrix[clf_1][clf_2])] += 1
        return self._labels[np.argmax(scores)]

    def _oversample(self, X, y):
        if self.preprocessing is None:
            return X, y

        if isinstance(self.preprocessing, str):
            if self.preprocessing not in OVO._allowed_preprocessing:
                raise ValueError("Unknown preprocessing: %s, expected to be one of %s."
                                 % (self.preprocessing, OVO._allowed_preprocessing))
            elif np.unique(y).size == 1:
                return X, y
            elif self.preprocessing == 'globalCS':
                gcs = GlobalCS()
                return gcs.fit_resample(X, y)
            elif self.preprocessing == 'SMOTE':
                return self._smote_oversample(X, y)
            elif self.preprocessing == 'SOUP':
                soup = SOUP()
                return soup.fit_resample(X, y)
        else:
            if not hasattr(self.preprocessing, 'fit_resample'):
                raise ValueError("Your resampler must implement fit_resample method")
            return self.preprocessing.fit_resample(X, y)

    def _smote_oversample(self, X, y):
        n_neighbors = min(3, min(np.unique(y, return_counts=True)[1]) - 1)
        if n_neighbors == 0:
            raise ValueError(
                'In order to use SMOTE preprocessing, the training set should contain at least 2 examples from each class')
        smote = SMOTE(k_neighbors=n_neighbors, random_state=42)
        return smote.fit_resample(X, y)

    def should_perform_oversampling(self, first_class, second_class):
        if self.oversample_between not in OVO._allowed_preprocessing_between:
            raise ValueError("Unknown strategy for oversampling: %s, expected to be one of %s."
                             % (self.oversample_between, OVO._allowed_preprocessing_between))
        elif self.oversample_between == 'all':
            return True
        elif self.oversample_between == 'maj-min':
            return (first_class in self._minority_classes and second_class not in self._minority_classes) or \
                   (second_class in self._minority_classes and first_class not in self._minority_classes)



import glob
from collections import OrderedDict, Counter
from pathlib import Path
from statistics import median

import numpy as np
import pandas as pd
from scipy.io import arff
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import Bunch


def construct_flat_2pc_df(X, y) -> pd.DataFrame:
    """
    This function takes two dimensional X and one dimensional y arrays, concatenates and returns them as data frame

    :param X:
        two dimensional numpy array
    :param y:
        one dimensional numpy array with labels
    :return:
        Data frame with 3 columns x1 x2 and y and with number of rows equal to number of rows in X
    """
    y = pd.DataFrame({'y': y})
    X_df = pd.DataFrame(data=X, columns=['x1', 'x2'])

    df = pd.concat([X_df, y], axis=1)

    return df


def get_project_root() -> Path:  # pragma no cover
    """Returns project root folder."""
    return Path(__file__).parent.parent.parent


def load_arff_dataset(path: str, one_hot_encode: bool = True, return_non_cat_length: bool = False):
    """
    Load and return the dataset saved in arff type file

    :param str path:
        location of dataset file
    :param bool one_hot_encode:
        flag, if true encodes categorical variables using OneHotEncoder
    :param bool return_non_cat_length:
        flag, if true returns the number of non categorical variables
    :returns:
        - ndarray X - dimensional numpy array where non categorical variables are stored in first columns followed by categorical variables
        - ndarray y - one dimensional numpy array with the classification target
        - bool non_cat_length - number of non categorical variables (only if return_non_cat_length=True)
    """
    data, meta = arff.loadarff(path)

    df = pd.DataFrame(data)
    y_index = len(df.columns) - 1
    y = df.pop(df.columns[y_index])

    le = LabelEncoder()
    y = le.fit_transform(y)

    categorical_feature_mask = df.dtypes == object

    categorical_cols = df.columns[categorical_feature_mask].tolist()
    non_categorical_cols = df.columns[~categorical_feature_mask].tolist()

    df[categorical_cols] = df[categorical_cols].replace({b'?': np.NaN})
    mode = df.mode().iloc[0]
    mean = df.filter(non_categorical_cols).mean()

    df[categorical_cols] = df.filter(categorical_cols).fillna(mode)
    df[non_categorical_cols] = df.filter(non_categorical_cols).fillna(mean)

    if one_hot_encode:
        X = pd.get_dummies(df, columns=categorical_cols)
    else:
        col_list = non_categorical_cols + categorical_cols
        X = df[col_list]

    if return_non_cat_length:
        return X.to_numpy(), y, len(non_categorical_cols)
    else:
        return X.to_numpy(), y


def load_datasets_arff(return_non_cat_length=False, dataset_paths=None):
    if dataset_paths is None:
        dataset_paths = glob.glob(f'{get_project_root()}/data/arff/*')

    datasets = OrderedDict()
    for path in sorted(dataset_paths):
        dataset_file = path.split('/')[-1]
        dataset_name = dataset_file.split('.')[0]
        if return_non_cat_length:
            X, y, cat_length = load_arff_dataset(path, return_non_cat_length=return_non_cat_length)
            datasets[dataset_name] = Bunch(data=X, target=y, non_cat_length=cat_length, DESCR=dataset_name)
        else:
            X, y = load_arff_dataset(path, return_non_cat_length=return_non_cat_length)
            datasets[dataset_name] = Bunch(data=X, target=y, DESCR=dataset_name)

    return datasets


def construct_maj_int_min(y: np.ndarray, strategy='median') -> OrderedDict:
    """
    This function creates dictionary with information which classes are minority or majority

    :param y:
        One dimensional numpy array that contains class labels
    :param strategy:
        The principle according to which the division into minority and majority classes will be determined:

        * 'median':
            A class whose size is equal to the median of the class sizes will be considered "intermediate"
        * 'average':
            The average class size will be calculated, all classes that are smaller will be considered as minority and
            the rest will be considered majority
    :return:
        dictionary with keys 'maj', 'int', 'min. The value for each key is a list containing the class labels belonging
        to the given group
    """
    class_sizes = Counter(y)

    if strategy == 'median':
        middle_size = median(list(class_sizes.values()))
    elif strategy == 'average':
        middle_size = np.mean(list(class_sizes.values()))
    else:
        raise ValueError(f'Unrecognized {strategy}. Only "median" and "average" are allowed.')

    maj_int_min = OrderedDict({
        'maj': list(),
        'int': list(),
        'min': list()
    })
    for class_label, class_size in class_sizes.items():
        if class_size == middle_size:
            class_group = 'int'
        elif class_size < middle_size:
            class_group = 'min'
        else:
            class_group = 'maj'

        maj_int_min[class_group].append(class_label)

    return maj_int_min



from collections import Counter
from copy import deepcopy
from math import sqrt

import numpy as np
from scipy.stats import multinomial
from sklearn.ensemble import BaggingClassifier
from sklearn.feature_selection import SelectKBest, chi2, f_classif
from sklearn.utils import resample
from sklearn.utils.random import sample_without_replacement


class MRBBagging(BaggingClassifier):
    """
    Multi-class Roughly Balanced Bagging (MRBBagging) is a generalization of MRBBagging for adapting to multiple
    minority classes.

    Reference:
    M. Lango, J. Stefanowski:
    Multi-class and feature selection extensions of RoughlyBalanced Bagging for imbalanced data.
    J. Intell Inf Syst (2018) 50: 97
    """

    def __init__(self, k, learning_algorithm, undersampling=True, feature_selection=False, random_fs=False,
                 half_features=True, random_state=None):
        """
        :param k:
            number of classifiers (multiplied by 3 when choosing feature selection)
        :param learning_algorithm:
            classifier to be used
        :param undersampling:
            (optional) boolean value to determine if undersampling or oversampling should be performed
        :param feature_selection:
            (optional) boolean value to determine if feature selection should be performed
        :param random_fs:
            (optional) boolean value to determine if feature selection should be all random (if False, chi^2, F test
            and random feature selection are performed)
        :param half_features:
            (optional) boolean value to determine if the number of features to be selected should be 50%
            (if False, it is set to the square root of the base number of features)
        :param random_state:
            (optional) the seed of the pseudo random number generator
        """
        super().__init__(random_state)
        assert learning_algorithm is not None, "Learning algorithm cannot be None"
        assert k > 0, "Number of classifiers must be > 0"
        self.classifiers, self.classes, self.classifier_classes = dict(), dict(), dict()
        self.feature_selection_methods = dict()
        self.k = k
        self.learning_algorithm = learning_algorithm
        self.undersampling = undersampling
        self.feature_selection = feature_selection
        self.all_random = random_fs
        self.half_features = half_features
        self.random_state = random_state

    def fit(self, x, y, **kwargs):
        """
        Build a MRBBagging ensemble of estimators from the training data.

        :param x:
            Two dimensional numpy array (number of samples x number of features) with float numbers.
        :param y:
            One dimensional numpy array with labels for rows in X.
        :return:
            self (object)
        """
        assert len(x) == len(y), "Not enough labels"

        classes, grouped_data = self._group_data(x, y)
        prob = [1 / len(classes)] * len(classes)
        self._set_classes_dict(classes)

        n = len(x)
        if self.undersampling:
            q = Counter(y)
            n = min(q.values()) * len(q)

        la_list = []

        if self.feature_selection:
            for i in range(3 * self.k):
                la_list.append(deepcopy(self.learning_algorithm))
            self._train_with_feature_selection(la_list, n, prob, classes, grouped_data)

        else:
            for i in range(self.k):
                la_list.append(deepcopy(self.learning_algorithm))
            self._train(la_list, n, prob, classes, grouped_data)

        return self

    def predict(self, data):
        """
        Predict classes for examples in data.

        :param data:
            Two dimensional numpy array (number of samples x number of features) with float numbers.
        """
        return self._select_classes(data)

    def _group_data(self, x, y):
        classes = set(y)
        self.classes = {key: value for (key, value) in enumerate(classes)}
        data = [[x[i], y[i]] for i in range(len(x))]
        grouped_data = dict()
        for cl in classes:
            assert cl is not None, "Missing class name"
            grouped_data[cl] = list(filter(lambda d: d[1] == cl, data))
        return classes, grouped_data

    def _resample(self, n, prob, classes, grouped_data):
        samples_no = multinomial.rvs(n=n, p=prob, random_state=self.random_state)
        subset_x, subset_y = [], []
        for no, j in enumerate(classes):
            data = grouped_data[j]
            resample_class = resample(data, replace=True, n_samples=samples_no[no], random_state=self.random_state)
            for sample in resample_class:
                subset_x.append(sample[0])
                subset_y.append(sample[1])
        return np.array(subset_x), np.array(subset_y)

    def _train(self, la_list, n, prob, classes, grouped_data):
        for i in range(len(la_list)):
            subset_x, subset_y = self._resample(n, prob, classes, grouped_data)

            subset_x = np.array(subset_x).astype(np.float)
            subset_y = np.array(subset_y).astype(np.float)

            self.classifiers[i] = la_list[i].fit(subset_x, subset_y)

    def _find_random_features(self, labels_no, features_no, subset_x):
        random_features_idx = sample_without_replacement(labels_no, features_no)
        random_features = self._get_features_array(subset_x, random_features_idx)
        return random_features, random_features_idx

    def _get_features_array(self, subset_x, random_features_idx):
        random_features = np.array(subset_x[:, random_features_idx[0]])
        for f in range(1, len(random_features_idx)):
            random_features = np.vstack((random_features, subset_x[:, random_features_idx[f]]))
        if random_features.ndim == 1:
            return random_features[:, np.newaxis]
        return random_features.T

    def _get_kbest_classifier(self, test, features_no, subset_x, subset_y):
        kBest_estimator = SelectKBest(test, k=features_no)
        subset = kBest_estimator.fit_transform(subset_x, subset_y)
        return subset, kBest_estimator

    def _train_with_feature_selection(self, la_list, n, prob, classes, grouped_data):
        for i in range(0, len(la_list), 3):
            subset_x, subset_y = self._resample(n, prob, classes, grouped_data)
            labels_no = len(subset_x[0])
            if self.half_features:
                features_no = int(labels_no / 2)
            else:
                features_no = int(sqrt(labels_no))

            subset_x = np.array(subset_x).astype(np.float)
            subset_y = np.array(subset_y).astype(np.float)

            if self.all_random:
                subset1, subset1_idx = self._find_random_features(labels_no, features_no, subset_x)
                subset2, subset2_idx = self._find_random_features(labels_no, features_no, subset_x)
                subset3, subset3_idx = self._find_random_features(labels_no, features_no, subset_x)

            else:
                subset1, subset1_idx = self._get_kbest_classifier(chi2, features_no, subset_x, subset_y)
                subset2, subset2_idx = self._get_kbest_classifier(f_classif, features_no, subset_x, subset_y)
                subset3, subset3_idx = self._find_random_features(labels_no, features_no, subset_x)

            self.feature_selection_methods[i] = subset1_idx
            self.feature_selection_methods[i + 1] = subset2_idx
            self.feature_selection_methods[i + 2] = subset3_idx

            self.classifiers[i] = la_list[i].fit(subset1, subset_y)
            self.classifiers[i + 1] = la_list[i + 1].fit(subset2, subset_y)
            self.classifiers[i + 2] = la_list[i + 2].fit(subset3, subset_y)

    def _set_classes_dict(self, classes):
        self.classifier_classes = dict(enumerate(classes))

    def _select_data(self, classifier_id, data):
        if self.feature_selection:
            if self.all_random:
                new_data = self._get_features_array(data, self.feature_selection_methods[classifier_id])
            else:
                if (classifier_id % 3) - 2 == 0:
                    new_data = self._get_features_array(data, self.feature_selection_methods[classifier_id])
                else:
                    new_data = self.feature_selection_methods[classifier_id].transform(data)
            return new_data
        return data

    def _count_votes(self, data):
        voting_matrix = np.zeros((len(data), len(self.classes)))
        for classifier_id in range(len(self.classifiers)):
            new_data = self._select_data(classifier_id, data)
            classes = self.classifiers[classifier_id].predict(new_data)
            probabilities = self.classifiers[classifier_id].predict_proba(new_data)
            for i, cl in enumerate(classes):
                idx = list(self.classifier_classes.keys())[list(self.classifier_classes.values()).index(int(cl))]
                voting_matrix[i][idx] += max(probabilities[i])
        return voting_matrix

    def _select_classes(self, data):
        voting_matrix = self._count_votes(data)
        selected_classes_ids = voting_matrix.argmax(axis=1)
        selected_classes = []
        for class_id in selected_classes_ids:
            selected_classes.append(self.classifier_classes[class_id])
        return selected_classes



from collections import Counter

import numpy as np
from imblearn.base import BaseSampler
from sklearn.neighbors import NearestNeighbors

from multi_imbalance.utils.array_util import (union, setdiff, contains)
from multi_imbalance.utils.data import construct_maj_int_min


class SPIDER3(BaseSampler):
    """
    SPIDER3 algorithm implementation for selective preprocessing of multi-class imbalanced data sets.

    Reference:
    Wojciechowski, S., Wilk, S., Stefanowski, J.: An Algorithm for Selective Preprocessing
    of Multi-class Imbalanced Data. Proceedings of the 10th International Conference
    on Computer Recognition Systems CORES 2017
    """

    def __init__(self, k, maj_int_min=None, cost=None):
        """
        :param k:
            Number of nearest neighbors considered while resampling.
        :param maj_int_min:
            Dict that contains lists of majority, intermediate and minority classes labels.
        :param cost:
            The cost matrix. An element c[i, j] of this matrix represents the cost associated with
            misclassifying an example from class i as class one from class j.
        """

        super().__init__()
        self._sampling_type = 'clean-sampling'
        self.k = k
        self.neigh_clf = NearestNeighbors(n_neighbors=self.k)
        self.maj_int_min = maj_int_min
        self.cost = cost
        self.AS, self.RS = np.array([]), np.array([])

    def _fit_resample(self, X, y):
        """
        Performs resampling

        :param X:
            Numpy array of examples that is the subject of resampling.
        :param y:
            Numpy array of labels corresponding to examples from X.
        :return:
            Resampled X along with accordingly modified labels, resampled y
        """
        self._initialize_algorithm(X, y)

        self.DS = np.append(X, y.reshape(y.shape[0], 1), axis=1)
        self._restart_perspective()
        self._calculate_weak_majority_examples()
        self._restore_perspective()
        self.DS = setdiff(self.DS, self.RS)
        int_classes, min_classes = self._sort_by_cardinality(y)

        for int_min_class in int_classes + min_classes:
            self.relabel(int_min_class)
            self.clean(int_min_class)
            self.amplify(int_min_class)

        self.DS = union(self.DS, self.AS)

        return self.DS[:, :-1], self.DS[:, -1]

    def _initialize_algorithm(self, X, y):
        if self.maj_int_min is None:
            self.maj_int_min = construct_maj_int_min(y)
        self.majority_classes = self.maj_int_min['maj']
        self.intermediate_classes = self.maj_int_min['int']
        self.minority_classes = self.maj_int_min['min']

        self.stds, self.means = [1] * X.shape[1], [0] * X.shape[1]
        if self.cost is None:
            self.cost = self._estimate_cost_matrix(y)

    @staticmethod
    def _estimate_cost_matrix(y):
        """
        Method that estimates cost matrix automatically. For example given imbalance ratios of 1:2:6, the estimated
        matrix will be:
        [0 1 1
        2 0 1
        6 3 0]
        :param y:
            labels
        :return:
            cost matrix
        """
        class_cardinality = Counter(y)
        classes = list(class_cardinality.keys())
        cost = np.ones([len(classes), len(classes)])
        for i, (c1, card1) in enumerate(class_cardinality.items()):
            for j, (c2, card2) in enumerate(class_cardinality.items()):
                if j > i:
                    cost[i, j] = 1
                else:
                    cost[i, j] = card1 / card2
        np.fill_diagonal(cost, 0)
        return cost

    def _sort_by_cardinality(self, y):
        class_cardinality = Counter(y)
        # to ensure looping over classes with decreasing cardinality.
        int_classes = sorted(self.intermediate_classes, key=lambda clazz: -class_cardinality[clazz])
        min_classes = sorted(self.minority_classes, key=lambda clazz: -class_cardinality[clazz])
        return int_classes, min_classes

    def amplify(self, int_min_class):
        self._restart_perspective()
        int_min_ds = self.DS[self.DS[:, -1] == int_min_class]
        for x in int_min_ds:
            self._amplify_nn(x)
        self._restore_perspective()

    def clean(self, int_min_class):
        self._restart_perspective()
        int_min_ds = self.DS[self.DS[:, -1] == int_min_class]
        int_min_as = self._calc_int_min_as(int_min_class)
        for x in union(int_min_ds, int_min_as):
            self._clean_nn(x)
        self._restore_perspective()

    def relabel(self, int_min_class):
        self._restart_perspective()
        int_min_ds = self.DS[self.DS[:, -1] == int_min_class]
        for x in int_min_ds:
            self._relabel_nn(x)
        self._restore_perspective()

    def _restart_perspective(self):
        """
        Performs normalization over resampled dataset.
        """
        for col in range(self._ds_as_rs_union().shape[1] - 1):
            self.stds[col] = self._ds_as_rs_union()[:, col].std()
            self.means[col] = self._ds_as_rs_union()[:, col].mean()

        for col in range(self._ds_as_rs_union().shape[1] - 1):
            if self.stds[col] == 0:
                self.stds[col] = 1e-6

        for dataset in [self.DS, self.RS, self.AS]:
            if dataset.shape[0] > 0:
                self._normalize(dataset)

    def _restore_perspective(self):
        """
        Denormalizes for further processing.
        """
        for dataset in [self.DS, self.RS, self.AS]:
            if dataset.shape[0] > 0:
                self._denormalize(dataset)

    def _normalize(self, dataset):
        for col in range(dataset.shape[1] - 1):
            dataset[:, col] = (dataset[:, col] - self.means[col]) / (4 * self.stds[col])

    def _denormalize(self, dataset):
        for col in range(dataset.shape[1] - 1):
            dataset[:, col] = dataset[:, col] * self.stds[col] * 4 + self.means[col]

    def _calc_int_min_as(self, int_min_class):
        """
        Helper method to calculate examples form AS that belong to int_min_class parameter class.
        :param int_min_class:
            The class name (intermediate or minority).
        :return:
            Examples from AS that are belong to int_min_class.
        """

        if self.AS.size != 0:
            int_min_as = self.AS[self.AS[:, -1] == int_min_class]
        else:
            int_min_as = np.array([])
        return int_min_as

    def _calculate_weak_majority_examples(self):
        """
        Calculates weak majority examples and appends them to the RS set.
        """

        for majority_class in self.majority_classes:
            majority_examples = self.DS[self.DS[:, -1] == majority_class]
            for x in majority_examples:
                if majority_class not in self._min_cost_classes(x, self.DS):
                    self.RS = union(self.RS, np.array([x]))

    def _min_cost_classes(self, x, DS):
        """
        Utility function that aims to identify minimum-cost classes, i.e. classes leading
        to the minimum cost after being (mis)classified as classes appearing in the neighborhood of x.

        :param x:
            Single observation
        :param DS:
            DS
        :return:
            List of classes associated with minimal cost of misclassification.
        """

        C = self.minority_classes + self.intermediate_classes + self.majority_classes
        vals = []
        kneighbors = self._knn(x, DS)

        for cj in C:
            s = 0
            for ci in C:
                s += ((kneighbors[:, -1] == ci).astype(int).sum() / self.k) * self.cost[C.index(ci), C.index(cj)]
            vals.append(s)
        C = np.array(C)
        vals = np.array(vals)
        vals = np.round(vals, 6)
        return C[vals == vals[np.argmin(vals)]]

    def _relabel_nn(self, x):
        """
        Performs relabeling in the nearest neighborhood of x.

        :param x:
            An observation.
        """
        nearest_neighbors = self._knn(x, self._ds_as_rs_union())
        for neighbor in nearest_neighbors:
            if contains(self.RS, neighbor) and self._class_of(neighbor) in self.majority_classes and self._class_of(
                    neighbor) in self._min_cost_classes(x, self._ds_as_rs_union()):
                self.RS = setdiff(self.RS, np.array([neighbor]))
                neighbor[-1] = x[-1]
                self.AS = union(self.AS, np.array([neighbor]))

    def _clean_nn(self, x):
        """
        Performs cleaning in the nearest neighborhood of x.

        :param x:
            Single observation.
        """
        nearest_neighbors = self._knn(x, self._ds_as_rs_union())
        for neighbor in nearest_neighbors:
            if self._class_of(neighbor) in self.majority_classes and \
                    self._class_of(neighbor) in self._min_cost_classes(x, self._ds_as_rs_union()):
                self.DS = setdiff(self.DS, np.array([neighbor]))
                self.RS = setdiff(self.RS, np.array([neighbor]))

    def _knn(self, x, DS):
        """
        Returns k nearest neighbors of x in DS that belong to c class if specified.

        :param x:
            Single observation
        :param DS:
            DS
        :param c:
            Class of neighbors that should be returned.
        :return:
            These neighbors from k nearest that belong to class c if specified. Otherwise all of them.
        """

        DS = setdiff(DS, np.array([x]))
        if DS.shape[0] < self.k:
            self.neigh_clf = NearestNeighbors(n_neighbors=DS.shape[0])
        else:
            self.neigh_clf = NearestNeighbors(n_neighbors=self.k)

        self.neigh_clf.fit(DS[:, :-1])

        within_radius = self.neigh_clf.radius_neighbors([x[:-1]], radius=
        self.neigh_clf.kneighbors([x[:-1]], return_distance=True)[0][0][-1] + 0.0001 *
        self.neigh_clf.kneighbors([x[:-1]], return_distance=True)[0][0][-1], return_distance=True)

        unique_distances = np.unique(sorted(within_radius[0][0]))
        all_distances = within_radius[0][0]
        all_indices = within_radius[1][0]
        indices = []
        for dist in unique_distances:
            if len(indices) < self.k:
                indices += (all_indices[all_distances == dist]).tolist()

        return DS[indices]

    def _amplify_nn(self, x):
        """
        Artificially amplifies example x by adding a copy of it to the AS.

        :param x:
            Single observation.
        """

        while self._class_of(x) not in self._min_cost_classes(x, self._ds_as_rs_union()):
            y = x.copy()
            self.AS = union(self.AS, np.asarray([y]))

    @staticmethod
    def _class_of(example):
        return example[-1]

    def _ds_as_rs_union(self):
        return union(self.DS, union(self.AS, self.RS))



import setuptools # pragma no cover

with open("README.md", "r") as fh:  # pragma no cover

    long_description = fh.read()

setuptools.setup(  # pragma no cover
    name="multi-imbalance",
    version="0.0.14",
    author="Damian Horna, Kamil Pluciński, Hanna Klimczak, Jacek Grycza",
    author_email="horna.damian@gmail.com",
    description="Python package for tackling multiclass imbalance problems.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/damian-horna/multi-imbalance",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    install_requires=[
        "numpy>=1.17.0",
        "scikit-learn>=0.22",
        "pandas>=0.25.1",
        "pytest>=5.1.2",
        "imbalanced-learn>=0.6.1",
        "coverage>=5.1",
        "pytest-cov>=2.8.1",
        "IPython>=7.13.0",
        "seaborn>=0.10.1",
        "matplotlib>=3.2.1",
    ]
)



import multiprocessing
from collections import Counter
from copy import deepcopy

import numpy as np
from sklearn.ensemble import BaggingClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils import resample

from multi_imbalance.resampling.soup import SOUP
from multi_imbalance.utils.array_util import setdiff


def fit_clf(args):
    return SOUPBagging.fit_classifier(args)


class SOUPBagging(BaggingClassifier):
    """
    Version of Bagging that applies SOUP in each classifier

    Reference:
    Lango, M., and Stefanowski, J. SOUP-Bagging: a new approach for multi-class
    imbalanced data classification. PP-RAI ’19: Polskie Porozumienie na Rzecz Sztucznej
    Inteligencji (2019).
    """

    def __init__(self, classifier=None, maj_int_min=None, n_classifiers=5):
        """
        :param classifier:
            Instance of classifier
        :param maj_int_min:
            dict {'maj': majority class labels, 'min': minority class labels}
        :param n_classifiers:
            number of classifiers
        """
        super().__init__()
        self.classifiers, self.clf_weights = list(), list()
        self.maj_int_min = maj_int_min
        self.num_core = multiprocessing.cpu_count()
        self.n_classifiers = n_classifiers
        self.classes = None
        for _ in range(n_classifiers):
            self.clf_weights.append(1)
            if classifier is not None:
                self.classifiers.append(deepcopy(classifier))
            else:
                self.classifiers.append(KNeighborsClassifier())

    @staticmethod
    def fit_classifier(args):
        clf, X, y, resampled, maj_int_min = args
        x_sampled, y_sampled = resampled

        out_of_bag = setdiff(np.hstack((X, y[:, np.newaxis])), np.hstack((x_sampled, y_sampled[:, np.newaxis])))
        x_out, y_out = out_of_bag[:, :-1], out_of_bag[:, -1].astype(int)

        x_resampled, y_resampled = SOUP(maj_int_min=maj_int_min).fit_resample(x_sampled, y_sampled)
        clf.fit(x_resampled, y_resampled)

        result = clf.predict_proba(x_out)
        class_sum_prob = np.sum(result, axis=0) + 0.001
        class_quantities = Counter(y_out)
        expected_sum_prob = np.array([class_quantities[i] for i in range(len(Counter(y)))])
        try:
            global_weights = expected_sum_prob / class_sum_prob
        except Exception:
            global_weights = np.ones(shape=len(Counter(y)))
            print(f'Exc {Counter(y)} {Counter(y_out)} {result.shape} {expected_sum_prob.shape} {class_sum_prob.shape}')
        return clf, global_weights

    def fit(self, X, y, **kwargs):
        """
        :param X:
            array-like, sparse matrix of shape = [n_samples, n_features] The training input samples.
        :param y:
            array-like, shape = [n_samples]. The target values (class labels).
        :param **kwargs:
            dict (optional)
        :return:
            self object
        """
        self.classes = np.unique(y)

        pool = multiprocessing.Pool(self.num_core)
        results = pool.map(fit_clf, [(clf, X, y, resample(X, y, stratify=y, random_state=i), self.maj_int_min)
                                     for i, clf in enumerate(self.classifiers)])
        pool.close()
        pool.join()
        for i, (clf, weights) in enumerate(results):
            self.classifiers[i] = clf
            self.clf_weights[i] = weights

        self.clf_weights = np.array(self.clf_weights)

    def predict(self, X, strategy: str = 'average'):
        """
        Predict class for X. The predicted class of an input sample is computed as the class with the highest
        sum of predicted probability.

        :param X:
            {array-like, sparse matrix} of shape = [n_samples, n_features]. The training input samples.
        :param strategy:
            * 'average' :
                takes max from average values in prediction
            * 'optimistic' :
                takes always best value of probability
            * 'pessimistic' :
                takes always the worst value of probability
            * 'mixed' :
                for minority classes takes optimistic strategy, and pessimistic for others. It requires maj_int_min
        :return:
            array of shape = [n_samples]. The predicted classes.
        """
        weights_sum = self.predict_proba(X)
        if strategy == 'average':
            p = np.sum(weights_sum, axis=0)
        elif strategy == 'optimistic':
            p = np.max(weights_sum, axis=0)
        elif strategy == 'pessimistic':
            p = np.min(weights_sum, axis=0)
        elif strategy == 'mixed':
            n_samples = X.shape[0]
            n_classes = self.classes.shape[0]
            p = np.zeros(shape=(n_samples, n_classes)) - 1

            for i in range(n_classes):
                two_dim_class_vector = weights_sum[:, :, i]  # [:,:,1] -> [classifiers x samples]
                if i in self.maj_int_min['min']:
                    squeeze_with_strategy = np.max(two_dim_class_vector, axis=0)
                else:
                    squeeze_with_strategy = np.min(two_dim_class_vector, axis=0)  # [1, n_samples, 1] -> [n_samples]
                p[:, i] = squeeze_with_strategy
            assert -1 not in p
        elif strategy == 'global':
            for i, weight in enumerate(self.clf_weights):
                weights_sum[i] *= weight
            p = np.sum(weights_sum, axis=0)
        else:
            raise KeyError(f'Incorrect strategy param: ${strategy}')

        y_result = np.argmax(p, axis=1)
        return y_result

    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        :param X:
            {array-like, sparse matrix} of shape = [n_samples, n_features]. The training input samples.
        :return:
            array of shape = [n_classifiers, n_samples, n_classes]. The class probabilities of the input samples.
        """
        n_samples = X.shape[0]
        n_classes = self.classes.shape[0]

        results = np.zeros(shape=(self.n_classifiers, n_samples, n_classes))

        for i, clf in enumerate(self.classifiers):
            results[i] = clf.predict_proba(X)

        return results



from collections import Counter

import numpy as np
from imblearn.over_sampling import SMOTE
from imblearn.base import BaseSampler


class StaticSMOTE(BaseSampler):
    """
    Static SMOTE implementation:

    Reference:
    Fernández-Navarro, F., Hervás-Martínez, C., Gutiérrez, P.A.: A dynamic over-sampling
    procedure based on sensitivity for multi-class problems. Pattern Recognit. 44, 1821–1833
    (2011)
    """
    def __init__(self):
        super().__init__()
        self._sampling_type = 'over-sampling'

    def _fit_resample(self, X, y):
        """
        Performs resampling

        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :param y:
            one dimensional numpy array with labels for rows in X
        :return:
            Resampled X and y as numpy arrays
        """
        cnt = Counter(y)
        min_class = min(cnt, key=cnt.get)
        X_original, y_original = X.copy(), y.copy()
        X_resampled, y_resampled = X.copy(), y.copy()

        M = len(list(cnt.keys()))
        for _ in range(M):
            sm = SMOTE(sampling_strategy={min_class: cnt[min_class] * 2})
            X_smote, y_smote = sm.fit_resample(X_original, y_original)
            X_added_examples = X_smote[y_smote == min_class][cnt[min_class]:, :]
            X_resampled = np.vstack([X_resampled, X_added_examples])
            y_resampled = np.hstack([y_resampled, y_smote[y_smote == min_class][cnt[min_class]:]])
            cnt = Counter(y_resampled)
            min_class = min(cnt, key=cnt.get)

        return X_resampled, y_resampled



from collections import Counter

import numpy as np
from imblearn.base import BaseSampler
from sklearn.decomposition import PCA
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_random_state

from multi_imbalance.utils.data import construct_maj_int_min


class MDO(BaseSampler):
    """
    Mahalanbois Distance Oversampling is an algorithm that oversamples all classes to a quantity of the major class.
    Samples for oversampling are chosen based on their k neighbours and new samples are created in random place but
    with the same Mahalanbois distance from the centre of class to chosen sample.

    """

    def __init__(self, k=5, k1_frac=.4, seed=0, prop=1, maj_int_min=None):
        """
        :param k:
            Number of neighbours considered during the neighbourhood analysis
        :param k1_frac:
            Ratio of the number of neighbours in the sample class to all neighbours in the neighbourhood.
            If the ratio is greater, the example will not be considered noise
        :param seed:
        :param prop:
            Oversampling ratio, if equal to one the class size after resampling will be equal to the size of
            the largest class
        :param maj_int_min:
            dict {'maj': majority class labels, 'min': minority class labels}
        """
        super().__init__()
        self._sampling_type = 'over-sampling'
        self.knn = NearestNeighbors(n_neighbors=k)
        self.k2 = k
        self.k1 = int(k * k1_frac)
        self.random_state = check_random_state(seed)
        self.X, self.y = None, None
        self.prop = prop
        self.class_balances = maj_int_min

    def _fit_resample(self, X, y):
        """
        :param X:
            two dimensional numpy array (number of samples x number of features) with float numbers
        :param y:
            one dimensional numpy array with labels for rows in X
        :return:
            resampled X, resampled y
        """
        if self.class_balances is None:
            self.class_balances = construct_maj_int_min(y)

        self.knn.fit(X)
        self.X, self.y = X, y

        oversampled_X, oversampled_y = self.X.copy(), self.y.copy()
        quantities = Counter(self.y)
        goal_quantity = int(max(list(quantities.values())))
        labels = list(set(self.y))
        minority_classes = self.class_balances['min']

        for class_label in labels:
            if minority_classes is not None and class_label not in minority_classes:
                continue

            chosen_minor_class_samples_to_oversample, weights = self._choose_samples(class_label)
            if len(chosen_minor_class_samples_to_oversample) == 0:
                continue

            oversampling_rate = int((goal_quantity - quantities[class_label]) * self.prop)
            if oversampling_rate > 0:
                if len(chosen_minor_class_samples_to_oversample) == 1:
                    oversampled_set = np.repeat(chosen_minor_class_samples_to_oversample, oversampling_rate, axis=0)
                else:
                    chosen_samples_features_mean = np.mean(chosen_minor_class_samples_to_oversample, axis=0)
                    zero_mean_samples = chosen_minor_class_samples_to_oversample - chosen_samples_features_mean

                    n_components = min(zero_mean_samples.shape)
                    pca = PCA(n_components=n_components).fit(zero_mean_samples)

                    uncorrelated_samples = pca.transform(zero_mean_samples)
                    variables_variance = np.diag(np.cov(uncorrelated_samples, rowvar=False))

                    oversampled_set = self._MDO_oversampling(uncorrelated_samples, variables_variance,
                                                             oversampling_rate,
                                                             weights)
                    oversampled_set = pca.inverse_transform(oversampled_set) + chosen_samples_features_mean

                oversampled_X = np.vstack((oversampled_X, oversampled_set))
                oversampled_y = np.hstack((oversampled_y, np.array([class_label] * oversampling_rate)))

        return oversampled_X, oversampled_y

    def _choose_samples(self, class_label):
        minor_class_indices = [i for i, value in enumerate(self.y) if value == class_label]
        minor_set = self.X[minor_class_indices]

        quantity_same_class_neighbours = self.calculate_same_class_neighbour_quantities(minor_set, class_label)
        chosen_minor_class_samples_to_oversample = minor_set[quantity_same_class_neighbours >= self.k1]

        weights = quantity_same_class_neighbours[quantity_same_class_neighbours >= self.k1] / self.k2
        weights_sum = np.sum(weights)

        if weights_sum != 0:
            weights /= np.sum(weights)
        elif len(weights) > 0:
            value = 1 / len(weights)
            weights += value

        return chosen_minor_class_samples_to_oversample, weights

    def _MDO_oversampling(self, T, v, oversampling_rate, weights):
        oversampled_set = list()
        V = np.clip(np.copy(v), a_min=0.001, a_max=None)
        for _ in range(oversampling_rate):
            idx = self.random_state.choice(np.arange(len(T)), p=weights)
            X = np.square(T[idx])
            a = np.sum(X / V)
            alpha_V = a * V
            alpha_V[alpha_V < 0.001] = 0.001

            s = 0
            features_vector = list()
            for alpha_V_j in alpha_V[:-1]:
                sqrt_avj = np.sqrt(alpha_V_j)
                r = self.random_state.uniform(low=-sqrt_avj, high=sqrt_avj)
                s += r ** 2 / alpha_V_j
                features_vector.append(r)

            last = (1 - s) * alpha_V[-1]
            last_feature = np.sqrt(last) if last > 0 else 0
            random_last_feature = self.random_state.choice([-last_feature, last_feature], 1)[0]

            features_vector.append(random_last_feature)
            oversampled_set.append(features_vector)

        return np.array(oversampled_set)

    def calculate_same_class_neighbour_quantities(self, S_minor, S_minor_label):
        minority_class_neighbours_indices = self.knn.kneighbors(S_minor, return_distance=False)
        quantity_with_same_label_in_neighbourhood = list()
        for i in range(len(S_minor)):
            sample_neighbours_indices = minority_class_neighbours_indices[i][1:]
            quantity_sample_neighbours_indices_with_same_label = sum(self.y[sample_neighbours_indices] == S_minor_label)
            quantity_with_same_label_in_neighbourhood.append(quantity_sample_neighbours_indices_with_same_label)
        return np.array(quantity_with_same_label_in_neighbourhood)
